__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Union

from simba.data_processors.timebins_movement_calculator import \
    TimeBinsMovementCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaCheckbox, SimBADropDown)
from simba.utils.checks import check_float
from simba.utils.enums import ConfigKey, Keys, Links
from simba.utils.errors import NoDataError


class MovementAnalysisTimeBinsPopUp(ConfigReader, PopUpMixin):
    """
    Tkinter pop-up for defining parameters when computing movements in time-bins.

    :example:
    >>> _ =  MovementAnalysisTimeBinsPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    >>> _ =  MovementAnalysisTimeBinsPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if len(self.outlier_corrected_paths) == 0:
            raise NoDataError(msg=f'No data files found in {self.outlier_corrected_dir} directory, cannot compute time-bins movement statistics.', source=self.__class__.__name__)
        PopUpMixin.__init__(self, title="TIME BINS: DISTANCE/VELOCITY", size=(400, 400), icon='run')
        self.animal_cnt_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT NUMBER OF ANIMALS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.animal_cnt_dropdown = SimBADropDown(parent=self.animal_cnt_frm, label="# OF ANIMALS", label_width=30, dropdown_width=20, value=1, dropdown_options=list(range(1, self.animal_cnt + 1)), command=self.create_bp_frm)
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW, padx=10, pady=10)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)

        self.plots_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="PLOTS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.plots_cb, self.plots_var = SimbaCheckbox(parent=self.plots_frm, txt='CREATE PLOTS', txt_img='plot', val=True)
        self.plots_frm.grid(row=1, column=0, sticky=NW, padx=10, pady=10)
        self.plots_cb.grid(row=0, column=0, sticky=NW)

        self.time_bin_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="TIME BIN", icon_name='timer_2', icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.time_bin_entry = Entry_Box(parent=self.time_bin_frm, fileDescription='TIME BIN SIZE (S): ', labelwidth=30, entry_box_width=20)
        self.time_bin_frm.grid(row=2, column=0, sticky=NW, padx=10, pady=10)
        self.time_bin_entry.grid(row=0, column=0, sticky=NW)
        self.create_bp_frm(animal_cnt=1)
        self.create_run_frm(run_function=self._run)

        self.main_frm.mainloop()

    def create_bp_frm(self, animal_cnt):
        if hasattr(self, "bp_frm"):
            self.bp_frm.destroy()
            for k, v in self.body_parts_dropdowns.items():
                v.destroy()

        self.bp_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT BODY-PARTS", icon_name='pose', icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.body_parts_dropdowns = {}
        for cnt, i in enumerate(range(int(animal_cnt))):
            self.body_parts_dropdowns[cnt] = SimBADropDown(parent=self.bp_frm, label=f"Animal {cnt+1}", label_width=30, dropdown_width=20, value=self.body_parts_lst[cnt], dropdown_options=self.body_parts_lst)
            self.body_parts_dropdowns[cnt].grid(row=cnt, column=0, sticky=NW)
        self.bp_frm.grid(row=3, column=0, sticky=NW, padx=10, pady=10)

    def _run(self):
        check_float(name="Time bin", value=str(self.time_bin_entry.entry_get), min_value=10e-6)
        self.config.set(ConfigKey.PROCESS_MOVEMENT_SETTINGS.value, ConfigKey.ROI_ANIMAL_CNT.value, str(self.animal_cnt_dropdown.getChoices()))
        body_parts = []
        for cnt, dropdown in self.body_parts_dropdowns.items():
            self.config.set(ConfigKey.PROCESS_MOVEMENT_SETTINGS.value, f"animal_{cnt + 1}_bp", str(dropdown.getChoices()))
            body_parts.append(dropdown.getChoices())
        self.update_config()
        time_bin_movement_analyzer = TimeBinsMovementCalculator(config_path=self.config_path,
                                                                bin_length=float(self.time_bin_entry.entry_get),
                                                                plots=self.plots_var.get(),
                                                                body_parts=body_parts)
        time_bin_movement_analyzer.run()
        time_bin_movement_analyzer.save()

#_ =  MovementAnalysisTimeBinsPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")



__author__ = "Simon Nilsson"

import os
import shutil
from typing import List, Optional, Union

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_int, check_str, check_valid_dataframe,
    check_valid_lst)
from simba.utils.data import create_color_palette, detect_bouts
from simba.utils.enums import Formats, Options
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.lookups import get_named_colors
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df

STYLE_WIDTH = 'width'
STYLE_HEIGHT = 'height'
STYLE_FONT_SIZE = 'font size'
STYLE_FONT_ROTATION = 'font size'

STYLE_ATTR = [STYLE_WIDTH, STYLE_HEIGHT, STYLE_FONT_SIZE, STYLE_FONT_ROTATION]

class GanttCreatorSingleProcess(ConfigReader, PlottingMixin):
    """
    Create gantt chart videos and/or images using a single core.

    .. note::
       `GitHub gantt tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#gantt-plot>`__.
       For improved run-time, see :meth:`simba.gantt_creator_mp.GanttCreatorMultiprocess` for multiprocess class.

    .. image:: _static/img/gantt_plot.png
       :width: 300
       :align: center


    :param str config_path: path to SimBA project config file in Configparser format.
    :param bool frame_setting: If True, creates individual frames.
    :param bool last_frm_setting: If True, creates single .png image representing entire video.
    :param bool video_setting: If True, creates videos
    :param dict style_attr: Attributes of gannt chart (size, font size, font rotation etc).
    :param List[str] files_found: File paths representing files with machine predictions e.g., ['project_folder/csv/machine_results/My_results.csv']

    :example:
    >>> style_attr = {'width': 640, 'height': 480, 'font size': 12, 'font rotation': 45}
    >>> gantt_creator = GanttCreatorSingleProcess(config_path='tests/test_data/multi_animal_dlc_two_c57/project_folder/project_config.ini', frame_setting=False, video_setting=True, files_found=['tests/test_data/multi_animal_dlc_two_c57/project_folder/csv/machine_results/Together_1.csv'])
    >>> gantt_creator.run()

    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_paths: List[Union[str, os.PathLike]],
                 width: int = 640,
                 height: int = 480,
                 font_size: int = 8,
                 font_rotation: int = 45,
                 palette: str = 'Set1',
                 frame_setting: Optional[bool] = False,
                 video_setting: Optional[bool] = False,
                 last_frm_setting: Optional[bool] = True):

        if ((frame_setting != True) and (video_setting != True) and (last_frm_setting != True)):
            raise NoSpecifiedOutputError(msg="SIMBA ERROR: Please select gantt videos, frames, and/or last frame.")
        check_file_exist_and_readable(file_path=config_path)
        check_int(value=width, min_value=1, name=f'{self.__class__.__name__} width')
        check_int(value=height, min_value=1, name=f'{self.__class__.__name__} height')
        check_int(value=font_size, min_value=1, name=f'{self.__class__.__name__} font_size')
        check_int(value=font_rotation, min_value=0, max_value=180, name=f'{self.__class__.__name__} font_rotation')
        check_valid_lst(data=data_paths, source=f'{self.__class__.__name__} data_paths', valid_dtypes=(str,), min_len=1)
        palettes = Options.PALETTE_OPTIONS_CATEGORICAL.value + Options.PALETTE_OPTIONS.value
        check_str(name=f'{self.__class__.__name__} palette', value=palette, options=palettes)
        for file_path in data_paths: check_file_exist_and_readable(file_path=file_path)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        self.clr_lst = create_color_palette(pallete_name=palette, increments=len(self.body_parts_lst)+1, as_int=True, as_rgb_ratio=True)
        if not os.path.exists(self.gantt_plot_dir): os.makedirs(self.gantt_plot_dir)
        self.frame_setting, self.video_setting, self.last_frm_setting = frame_setting, video_setting, last_frm_setting
        self.width, self.height, self.font_size, self.font_rotation = width, height, font_size, font_rotation
        self.data_paths = data_paths
        self.colours = get_named_colors()
        self.colour_tuple_x = list(np.arange(3.5, 203.5, 5))

    def run(self):
        print(f"Processing gantt visualizations for {len(self.data_paths)} video(s)...")
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            _, self.video_name, _ = get_fn_ext(file_path)
            self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
            check_valid_dataframe(df=self.data_df, source=file_path, required_fields=self.clf_names)
            self.video_info_settings, _, self.fps = self.read_video_info(video_name=self.video_name)
            self.bouts_df = detect_bouts(data_df=self.data_df, target_lst=self.clf_names, fps=self.fps)
            if self.frame_setting:
                self.save_frame_folder_dir = os.path.join(self.gantt_plot_dir, self.video_name)
                if os.path.exists(self.save_frame_folder_dir):
                    shutil.rmtree(self.save_frame_folder_dir)
                os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.save_video_path = os.path.join(self.gantt_plot_dir, f"{self.video_name}.mp4")
                self.writer = cv2.VideoWriter(self.save_video_path, self.fourcc, self.fps, (self.width, self.height))
            if self.last_frm_setting:
                self.make_gantt_plot(x_length=len(self.data_df),
                                     bouts_df=self.bouts_df,
                                     clf_names=self.clf_names,
                                     fps=self.fps,
                                     width = self.width,
                                     height=self.height,
                                     font_size=self.font_size,
                                     font_rotation=self.font_rotation,
                                     video_name=self.video_name,
                                     save_path=os.path.join(self.gantt_plot_dir, f"{self.video_name }_final_image.png"),
                                     palette=self.clr_lst)

            if self.frame_setting or self.video_setting:
                for image_cnt, k in enumerate(range(len(self.data_df))):
                    frm_bout_df = self.bouts_df.loc[self.bouts_df["End_frame"] <= k]
                    plot = self.make_gantt_plot(x_length=k+1,
                                                bouts_df=frm_bout_df,
                                                clf_names=self.clf_names,
                                                fps=self.fps,
                                                width=self.width,
                                                height=self.height,
                                                font_size=self.font_size,
                                                font_rotation=self.font_rotation,
                                                video_name=self.video_name,
                                                palette=self.clr_lst,
                                                save_path=None)
                    if self.frame_setting:
                        frame_save_path = os.path.join(self.save_frame_folder_dir, f"{k}.png")
                        cv2.imwrite(frame_save_path, plot)
                    if self.video_setting:
                        self.writer.write(plot)
                    print(f"Gantt frame: {image_cnt + 1} / {len(self.data_df)}. Video: {self.video_name} ({file_cnt + 1}/{len(self.data_paths)})")
                if self.video_setting:
                    self.writer.release()
                    print(f"Gantt for video {self.video_name} saved...")
            self.timer.stop_timer()
            stdout_success(msg=f"All gantt visualizations created in {self.gantt_plot_dir} directory", elapsed_time=self.timer.elapsed_time_str)


# test = GanttCreatorSingleProcess(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                                 frame_setting=False,
#                                 video_setting=False,
#                                 data_paths=[r"C:\troubleshooting\mitra\project_folder\csv\machine_results\592_MA147_Gq_CNO_0515.csv"],
#                                 last_frm_setting=True,
#                                 width=640,
#                                 height= 480,
#                                 font_size=10,
#                                 font_rotation=45,
#                                 palette='Set1')
# test.run()




# style_attr = {'width': 640, 'height': 480, 'font size': 12, 'font rotation': 45}
# test = GanttCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini',
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  last_frm_setting=True,
#                                  style_attr=style_attr,
#                                  files_found=['/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/csv/outlier_corrected_movement_location/Testing_Video_3.csv'])
# test.create_gannt()


# style_attr = {'width': 640, 'height': 480, 'font size': 12, 'font rotation': 45}
# test = GanttCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  last_frm_setting=True,
#                                  style_attr=style_attr,
#                                  files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/targets_inserted/Together_1.csv'])
# test.run()



__author__ = "Tzuk Polinsky"

import os
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.directing_animals_to_bodypart_visualizer import \
    DirectingAnimalsToBodyPartVisualizer
from simba.plotting.directing_animals_visualizer import \
    DirectingOtherAnimalsVisualizer
from simba.plotting.directing_animals_visualizer_mp import \
    DirectingOtherAnimalsVisualizerMultiprocess
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.enums import Formats, Keys, Links, Paths
from simba.utils.errors import AnimalNumberError
from simba.utils.read_write import get_file_name_info_in_directory


class DirectingAnimalToBodyPartVisualizerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        PopUpMixin.__init__(self, title="CREATE ANIMAL DIRECTION TO BODY PART VIDEOS", icon='eye')
        self.color_lst = list(self.colors_dict.keys())
        self.color_lst.insert(0, "Random")
        self.size_lst = list(range(1, 11))
        self.data_path = os.path.join(self.project_path, Paths.OUTLIER_CORRECTED.value)
        self.files_found_dict = get_file_name_info_in_directory(directory=self.data_path, file_type=self.file_type)

        self.show_pose_var = BooleanVar(value=True)
        self.merge_directionality_lines_var = BooleanVar(value=False)
        self.multiprocess_var = BooleanVar(value=False)

        self.style_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm,header="STYLE SETTINGS",icon_name=Keys.DOCUMENTATION.value,icon_link=Links.DIRECTING_ANIMALS_PLOTS.value,)
        self.show_pose_cb = Checkbutton(self.style_settings_frm,font=Formats.FONT_REGULAR.value,text="Show pose-estimated body-parts",variable=self.show_pose_var)
        self.merge_directionality_lines_cb = Checkbutton(self.style_settings_frm,font=Formats.FONT_REGULAR.value,text="Polyfill direction lines",variable=self.merge_directionality_lines_var,)
        self.direction_clr_dropdown = DropDownMenu(self.style_settings_frm, "Direction color:", self.color_lst, "16")
        self.pose_size_dropdown = DropDownMenu(self.style_settings_frm, "Pose circle size:", self.size_lst, "16")
        self.line_thickness = DropDownMenu(self.style_settings_frm, "Line thickness:", self.size_lst, "16")
        self.line_thickness.setChoices(choice=4)
        self.pose_size_dropdown.setChoices(choice=3)
        self.direction_clr_dropdown.setChoices(choice="Random")
        # multiprocess_cb = Checkbutton(
        #     self.style_settings_frm,
        #     text="Multi-process (faster)",
        #     variable=self.multiprocess_var,
        #     command=lambda: self.enable_dropdown_from_checkbox(
        #         check_box_var=self.multiprocess_var,
        #         dropdown_menus=[self.multiprocess_dropdown],
        #     ),
        # )
        # self.multiprocess_dropdown = DropDownMenu(
        #     self.style_settings_frm, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        # )
        # self.multiprocess_dropdown.setChoices(2)
        # self.multiprocess_dropdown.disable()

        self.run_frm = LabelFrame( self.main_frm, text="RUN", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black")
        self.run_single_video_frm = LabelFrame( self.run_frm, text="SINGLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black",)
        self.run_single_video_btn = Button( self.run_single_video_frm, font=Formats.FONT_REGULAR.value, text="Create single video", fg="blue", command=lambda: self.__create_directionality_plots(multiple_videos=False),)
        self.single_video_dropdown = DropDownMenu(self.run_single_video_frm, "Video:", list(self.files_found_dict.keys()), "12")
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])
        self.run_multiple_videos = LabelFrame( self.run_frm, text="MULTIPLE VIDEO", font=Formats.FONT_HEADER.value, pady=5, padx=5, fg="black",)
        self.run_multiple_video_btn = Button(
            self.run_multiple_videos,
            font=Formats.FONT_REGULAR.value,
            text="Create multiple videos ({} video(s) found)".format(
                str(len(list(self.files_found_dict.keys())))
            ),
            fg="blue",
            command=lambda: self.__create_directionality_plots(multiple_videos=True),
        )

        self.style_settings_frm.grid(row=0, column=0, sticky=NW)
        self.show_pose_cb.grid(row=0, column=0, sticky=NW)
        self.merge_directionality_lines_cb.grid(row=2, column=0, sticky=NW)
        self.direction_clr_dropdown.grid(row=3, column=0, sticky=NW)
        self.pose_size_dropdown.grid(row=4, column=0, sticky=NW)
        self.line_thickness.grid(row=5, column=0, sticky=NW)
        # multiprocess_cb.grid(row=6, column=0, sticky=NW)
        # self.multiprocess_dropdown.grid(row=6, column=1, sticky=NW)

        self.run_frm.grid(row=1, column=0, sticky=NW)
        self.run_single_video_frm.grid(row=0, column=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)
        self.run_multiple_videos.grid(row=1, column=0, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, column=0, sticky=NW)

    def __create_directionality_plots(self, multiple_videos: bool):
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        style_attr = {
            "Show_pose": self.show_pose_var.get(),
            "Pose_circle_size": int(self.pose_size_dropdown.getChoices()),
            "Direction_color": self.direction_clr_dropdown.getChoices(),
            "Direction_thickness": int(self.line_thickness.getChoices()),
            "Polyfill": self.merge_directionality_lines_var.get(),
        }

        for data_path in data_paths:
            # if not self.multiprocess_var.get():
            directing_other_animal_visualizer = DirectingAnimalsToBodyPartVisualizer(
                config_path=self.config_path,
                data_path=data_path,
                style_attr=style_attr,
            )
            # else:
            #     directing_other_animal_visualizer = (
            #         DirectingOtherAnimalsVisualizerMultiprocess(
            #             config_path=self.config_path,
            #             data_path=data_path,
            #             style_attr=style_attr,
            #             core_cnt=int(self.multiprocess_dropdown.getChoices()),
            #         )
            #     )
            directing_other_animal_visualizer.run()



import os
import random
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.third_party_label_appenders.converters import create_yolo_yaml
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int, check_str,
                                check_valid_boolean, check_valid_dataframe,
                                check_valid_tuple,
                                check_video_and_data_frm_count_align)
from simba.utils.enums import Formats, Options
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data, read_df,
                                    read_frm_of_video)
from simba.utils.warnings import NoDataFoundWarning


class SimBA2YoloSegmentation(ConfigReader):
    """
    Convert pose estimation data from a SimBA project into the YOLO keypoint format, including frame sampling,
    image-label pair creation, bounding box computation, and train/validation splitting.

    .. note::
       For creating the ``flip_idx``, see :func:`simba.third_party_label_appenders.converters.get_yolo_keypoint_flip_idx`.
       For creating the ``bp_id_idx``, see :func:`simba.third_party_label_appenders.converters.get_yolo_keypoint_bp_id_idx`

    :param Union[str, os.PathLike] config_path: Path to the SimBA project `.ini` configuration file.
    :param Union[str, os.PathLike] save_dir: Directory where YOLO-formatted data will be saved. Subdirectories for images/labels (train/val) are created.
    :param Optional[Union[str, os.PathLike] data_dir: Optional directory containing outlier-corrected SimBA pose estimation data. If None, uses path from config.
    :param float train_size: Proportion of samples to allocate to the training set (range 0.1–0.99). Remaining samples go to validation.
    :param bool verbose: If True, prints progress updates to the console.
    :param bool greyscale: If True, saves extracted video frames in greyscale. Otherwise, saves in color.
    :param float padding: Padding added around the bounding box (as a proportion of image dimensions, range 0.0–1.0). Useful if animal body-parts are in a "line".
    :param Tuple[int, ...] flip_idx: Tuple defining symmetric keypoint indices for horizontal flipping. Used to write the `map.yaml` file. If None, then attempt to infer.
    :param Dict[int, str] names: Dictionary mapping instance IDs to class names. Used in annotation labels and `map.yaml`.
    :param Optional[int] sample_size: If specified, limits the number of randomly sampled frames per video. If None, all frames are used.
    :param Optional[Dict[int, Union[Tuple[int], List[int]]]] bp_id_idx: Optional mapping of instance IDs to keypoint index groups, allowing support for multiple animals per frame. Must match keys in `map_dict`.
    :param Optional[str] single_id: If the data contains pose-estimation for multiple indivisuals, but you want to treat it as examples of a single individual, pass the name of the single individual. Defaults to None, and the YOLO data will be formatted to the number of objects which the H5 data contains.
    :return: None. Saves YOLO-formatted images and annotations to disk in the `save_dir` location.

    :example:
    >>> SAVE_DIR = r'D:\troubleshooting\mitra\mitra_yolo'
    >>> CONFIG_PATH = r"C:\troubleshooting\mitra\project_folder\project_config.ini"
    >>> runner = SimBA2Yolo(config_path=CONFIG_PATH, save_dir=SAVE_DIR, sample_size=10, verbose=True)
    >>> runner.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 train_size: float = 0.7,
                 verbose: bool = False,
                 greyscale: bool = False,
                 clahe: bool = False,
                 padding: int = 0,
                 threshold: float = 0.00,
                 sample_size: Optional[int] = None,
                 single_id: Optional[str] = None) -> None:


        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose')
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale')
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe')
        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False, create_logger=False)
        if padding is not None: check_int(name=f'{self.__class__.__name__} padding', value=padding, min_value=0, raise_error=True)
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, max_value=1.0, min_value=0.0)
        check_if_dir_exists(in_dir=save_dir)
        self.img_dir, self.lbl_dir = os.path.join(save_dir, 'images'), os.path.join(save_dir, 'labels')
        self.img_train_dir, self.img_val_dir = os.path.join(self.img_dir, 'train'), os.path.join(self.img_dir, 'val')
        self.lbl_train_dir, self.lb_val_dir = os.path.join(self.lbl_dir, 'train'), os.path.join(self.lbl_dir, 'val')
        create_directory(paths=[self.img_train_dir, self.img_val_dir, self.lbl_train_dir, self.lb_val_dir], overwrite=False)
        self.map_path = os.path.join(save_dir, 'map.yaml')
        if single_id is not None: check_str(name=f'{self.__class__.__name__} single_id', value=single_id, raise_error=True)
        if sample_size is not None: check_int(name=f'{self.__class__.__name__} sample', value=sample_size, min_value=1)
        self.config = ConfigReader(config_path=config_path)
        if data_dir is not None:
            check_if_dir_exists(in_dir=data_dir, source=f'{self.__class__.__name__} data_dir')
        else:
            data_dir = self.outlier_corrected_dir
        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=[f'.{self.file_type}'], raise_error=True, as_dict=True)
        self.video_paths = find_files_of_filetypes_in_directory(directory=self.video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True, as_dict=True)
        missing_videos = [x for x in self.data_paths.keys() if x not in self.video_paths.keys()]
        if len(missing_videos) > 0:
            NoDataFoundWarning(msg=f'Data files {missing_videos} do not have corresponding videos in the {self.video_dir} directory', source=self.__class__.__name__)
        self.data_w_video = [x for x in self.data_paths.keys() if x in self.video_paths.keys()]
        if len(self.data_w_video) == 0:
            raise NoFilesFoundError(msg=f'None of the data files in {data_dir} have matching videos in the {self.video_dir} directory', source=self.__class__.__name__)
        self.sample_size, self.train_size, self.verbose, self.save_dir = sample_size, train_size, verbose, save_dir
        self.greyscale, self.padding = greyscale, padding
        self.clahe, self.threshold, self.single_id = clahe, threshold, single_id
        if self.single_id is None:
            self.map_ids = {v: k for k, v in enumerate(self.animal_bp_dict.keys())}
        else:
            self.map_ids = {self.single_id: '0'}

    def run(self):
        timer, annotations = SimbaTimer(start=True), []
        for file_cnt, video_name in enumerate(self.data_w_video):
            data = read_df(file_path=self.data_paths[video_name], file_type=self.config.file_type)
            check_valid_dataframe(df=data, source=f'{self.__class__.__name__} {self.data_paths[video_name]}', valid_dtypes=Formats.NUMERIC_DTYPES.value)
            video_path = self.video_paths[video_name]
            check_video_and_data_frm_count_align(video=video_path, data=data, name=self.data_paths[video_name], raise_error=True)
            p_data = data[data.columns[list(data.columns.str.endswith('_p'))]]
            data = data.loc[:, ~data.columns.str.endswith('_p')].reset_index(drop=True)
            data = data.iloc[(p_data[(p_data > self.threshold).all(axis=1)].index)]
            data['video'], frm_cnt = video_name, len(data)
            if self.sample_size is None:
                video_sample_idx = list(range(0, frm_cnt))
            else:
                video_sample_idx = list(range(0, frm_cnt)) if self.sample_size > frm_cnt else random.sample(list(range(0, frm_cnt)), self.sample_size)
            annotations.append(data.iloc[video_sample_idx].reset_index(drop=False))

        annotations = pd.concat(annotations, axis=0).reset_index(drop=True)
        video_names = annotations.pop('video').reset_index(drop=True).values
        train_idx = random.sample(list(annotations['index']), int(len(annotations) * self.train_size))


        for cnt, (idx, idx_data) in enumerate(annotations.iterrows()):
            vid_path = self.video_paths[video_names[cnt]]
            video_meta = get_video_meta_data(video_path=vid_path)
            if self.verbose:
                print(f'Processing annotation {cnt + 1}/{len(annotations)} from SimBA to YOLO segmentation.. ({video_meta["video_name"]})...')
            img = read_frm_of_video(video_path=vid_path, frame_index=idx_data['index'], greyscale=self.greyscale, clahe=self.clahe)
            img_lbl = ''
            img_name = f'{video_meta["video_name"]}.{idx}'
            if idx_data['index'] in train_idx:
                label_save_path, img_save_path = os.path.join(self.lbl_train_dir, f'{img_name}.txt'), os.path.join(self.img_train_dir, f'{img_name}.png')
            else:
                label_save_path, img_save_path = os.path.join(self.lb_val_dir, f'{img_name}.txt'), os.path.join(self.img_val_dir, f'{img_name}.png')
            for animal_cnt, (animal_name, bps_data) in enumerate(self.animal_bp_dict.items()):
                cols = [item for group in zip(bps_data['X_bps'], bps_data['Y_bps']) for item in group]
                keypoints = annotations.loc[idx, cols].values.reshape(-1, 2).astype(np.int32)
                if self.padding is not None and self.padding > 0:
                    keypoints = keypoints.reshape(-1, keypoints.shape[0], 2).astype(np.int32)
                    keypoints = GeometryMixin().bodyparts_to_polygon(data=keypoints, parallel_offset=self.padding)[0]
                    keypoints = np.array(keypoints.exterior.coords).astype(np.int32)
                mask = ~(np.all(keypoints == 0, axis=1) | np.all(keypoints < 0, axis=1) | np.any(~np.isfinite(keypoints), axis=1))
                keypoints = keypoints[mask]
                if keypoints.shape[0] > 2:
                    instance_str = f'{animal_cnt} ' if self.single_id is None else '0 '
                    pts = np.unique(keypoints, axis=0)
                    center = np.mean(pts, axis=0)
                    angles = np.arctan2(pts[:, 1] - center[1], pts[:, 0] - center[0])
                    seg_arr = pts[np.argsort(angles)]
                    seg_arr_x, seg_arr_y = np.clip(seg_arr[:, 0].flatten() / video_meta['width'], 0, 1), np.clip(seg_arr[:, 1].flatten() / video_meta['height'], 0, 1)
                    kps = list(np.column_stack((seg_arr_x, seg_arr_y)).flatten())
                    instance_str += ' '.join(str(x) for x in kps).strip() + '\n'
                    img_lbl += instance_str
            if img_lbl:
                with open(label_save_path, mode='wt', encoding='utf-8') as f:
                    f.write(img_lbl)
                cv2.imwrite(img_save_path, img)

        create_yolo_yaml(path=self.save_dir, train_path=self.img_train_dir, val_path=self.img_val_dir, names=self.map_ids, save_path=self.map_path)
        timer.stop_timer()
        if self.verbose: stdout_success(msg=f'Labelme to YOLO conversion complete. Data saved in directory {self.save_dir}.', elapsed_time=timer.elapsed_time_str)

# SAVE_DIR = r'D:\troubleshooting\mitra\mitra_yolo_seg'
# CONFIG_PATH = r"C:\troubleshooting\mitra\project_folder\project_config.ini"
# runner = SimBA2YoloSegmentation(config_path=CONFIG_PATH, save_dir=SAVE_DIR, sample_size=250, verbose=True, padding=None)
# runner.run()


__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_dir_exists, check_valid_boolean,
    check_valid_dataframe, check_valid_lst)
from simba.utils.data import detect_bouts
from simba.utils.enums import TagNames
from simba.utils.errors import NoChoosenMeasurementError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df)
from simba.utils.warnings import NoDataFoundWarning

FIRST_OCCURRENCE = "First occurrence (s)"
EVENT_COUNT = "Event count"
TOTAL_EVENT_DURATION = "Total event duration (s)"
MEAN_EVENT_DURATION = "Mean event duration (s)"
MEDIAN_EVENT_DURATION = "Median event duration (s)"
MEAN_EVENT_INTERVAL = "Mean event interval (s)"
MEDIAN_EVENT_INTERVAL = "Median event interval (s)"
FRAME_COUNT = 'Frame count'
VIDEO_LENGTH = "Video length (s)"
MEASUREMENT_NAMES = [FIRST_OCCURRENCE, EVENT_COUNT, TOTAL_EVENT_DURATION, MEAN_EVENT_DURATION, MEDIAN_EVENT_DURATION, MEAN_EVENT_INTERVAL, MEDIAN_EVENT_INTERVAL]







class AggregateClfCalculator(ConfigReader):
    """
    Compute aggregate descriptive statistics from classification data.

    .. note::
       `GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results>`__.
       `Example expected ouput file <https://github.com/sgoldenlab/simba/blob/master/misc/detailed_bout_data_summary_20231011091832.csv>`__.


    :param str config_path: path to SimBA project config file in Configparser format
    :param List[str] data_measures: Aggregate statistics measures to calculate. OPTIONS: ['Bout count', 'Total event duration (s)', 'Mean event bout duration (s)', 'Median event bout duration (s)', 'First event occurrence (s)', 'Mean event bout interval duration (s)', 'Median event bout interval duration (s)']. If None, then all measures are calculated.
    :param List[str] classifiers: Classifiers to calculate aggregate statistics for. E.g.,: ['Attack', 'Sniffing']
    :param Optional[List[str]] video_meta_data: Video metadata to include in the output. Options: 'Frame count', 'Video length (s)'.
    :param bool detailed_bout_data: If True, save detailed data for each bout in each video (start frame, end frame, bout time etc.)
    :param bool transpose: If True, then one video per row. Else, one meassure per row. Default: False.
    :param Optional[Union[str, os.PathLike]] data_dir: Directory location of the data files. If None, the the ``project_folder/csv/machine_results`` directory is used.


    :example:
    >>> clf_log_creator = AggregateClfCalculator(config_path="MyConfigPath", data_measures=['Bout count', 'Total event duration (s)'], classifiers=['Attack', 'Sniffing'])
    >>> clf_log_creator.run()
    >>> clf_log_creator.save()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 classifiers: List[str],
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 detailed_bout_data: bool = False,
                 transpose: bool = False,
                 first_occurrence: bool = True,
                 event_count: bool = True,
                 total_event_duration: bool = True,
                 mean_event_duration: bool = True,
                 median_event_duration: bool = True,
                 mean_interval_duration: bool = True,
                 median_interval_duration: bool = True,
                 frame_count: bool = False,
                 video_length: bool = False):

        super().__init__(config_path=config_path)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        check_valid_lst(data=classifiers, source=f'{self.__class__.__name__} classifiers', min_len=1, valid_dtypes=(str,), valid_values=self.clf_names)
        if data_dir is None:
            data_dir = self.machine_results_dir
        else:
            check_if_dir_exists(in_dir=data_dir)
        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=[f'.{self.file_type}'], raise_error=True)

        self.measurements = []
        for i, j  in zip([first_occurrence, event_count, total_event_duration, mean_event_duration, median_event_duration, mean_interval_duration, median_interval_duration], MEASUREMENT_NAMES):
            check_valid_boolean(value=i, source=f'{self.__class__.__name__} {j}', raise_error=True)
            self.measurements.append(j)

        check_valid_boolean(value=frame_count, source=f'{self.__class__.__name__} frame_count', raise_error=True)
        check_valid_boolean(value=video_length, source=f'{self.__class__.__name__} video_length', raise_error=True)
        check_valid_boolean(value=transpose, source=f'{self.__class__.__name__} transpose', raise_error=True)
        check_valid_boolean(value=detailed_bout_data, source=f'{self.__class__.__name__} detailed_bout_data', raise_error=True)
        if not any([first_occurrence, event_count, total_event_duration, mean_event_duration, median_event_duration, mean_interval_duration, median_interval_duration]):
            raise NoChoosenMeasurementError(source=self.__class__.__name__)

        self.classifiers, self.detailed_bout_data, self.transpose = classifiers, detailed_bout_data, transpose
        self.first_occurrence = first_occurrence
        self.event_count = event_count
        self.total_event_duration = total_event_duration
        self.mean_event_duration = mean_event_duration
        self.median_event_duration = median_event_duration
        self.mean_interval_duration = mean_interval_duration
        self.median_interval_duration = median_interval_duration
        self.frame_count = frame_count
        self.video_length = video_length
        self.save_path = os.path.join(self.logs_path, f"data_summary_{self.datetime}.csv")
        self.detailed_save_path = os.path.join(self.logs_path, f"detailed_bout_data_summary_{self.datetime}.csv")

    def run(self):
        self.results_df, self.bouts_df_lst = pd.DataFrame(), []
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.machine_results_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            _, file_name, _ = get_fn_ext(file_path)
            print(f"Analyzing classifier descriptive statistics for video {file_name} ({file_cnt+1}/{len(self.machine_results_paths)})...")
            _, _, fps = self.read_video_info(video_name=file_name)
            check_file_exist_and_readable(file_path)
            data_df = read_df(file_path, self.file_type)
            check_valid_dataframe(df=data_df, required_fields=self.classifiers, source=file_path)
            bouts_df = detect_bouts(data_df=data_df, target_lst=self.classifiers, fps=fps)
            if self.detailed_bout_data and (len(bouts_df) > 0):
                bouts_df_for_detailes = deepcopy(bouts_df)
                bouts_df_for_detailes.insert(loc=0, column="Video", value=file_name)
                self.bouts_df_lst.append(bouts_df_for_detailes)
            bouts_df["Shifted start"] = bouts_df["Start_time"].shift(-1)
            bouts_df["Interval duration"] = (bouts_df["Shifted start"] - bouts_df["End Time"])
            for clf in self.classifiers:
                clf_results_dict = {}
                clf_data = bouts_df.loc[bouts_df["Event"] == clf]
                if len(clf_data) > 0:
                    clf_results_dict[FIRST_OCCURRENCE] = round(clf_data["Start_time"].min(), 3)
                    clf_results_dict[EVENT_COUNT] = len(clf_data)
                    clf_results_dict[TOTAL_EVENT_DURATION] = round(clf_data["Bout_time"].sum(), 3)
                    clf_results_dict[MEAN_EVENT_DURATION] = round(clf_data["Bout_time"].mean(), 3)
                    clf_results_dict[MEDIAN_EVENT_DURATION] = round(clf_data["Bout_time"].median(), 3)
                else:
                    clf_results_dict[FIRST_OCCURRENCE] = None
                    clf_results_dict[EVENT_COUNT] = None
                    clf_results_dict[TOTAL_EVENT_DURATION] = None
                    clf_results_dict[MEAN_EVENT_DURATION] = None
                    clf_results_dict[MEDIAN_EVENT_DURATION] = None
                if len(clf_data) > 1:
                    interval_df = clf_data[:-1].copy()
                    clf_results_dict[MEAN_EVENT_INTERVAL] = round(interval_df["Interval duration"].mean(), 3)
                    clf_results_dict[MEDIAN_EVENT_INTERVAL] = round(interval_df["Interval duration"].median(), 3)
                else:
                    clf_results_dict[MEAN_EVENT_INTERVAL] = None
                    clf_results_dict[MEDIAN_EVENT_INTERVAL] = None
                if self.frame_count:
                        clf_results_dict[FRAME_COUNT] = len(data_df)
                if self.video_length:
                    clf_results_dict["Video length (s)"] = round(len(data_df) / fps, 3)
                video_clf_pd = (pd.DataFrame.from_dict(clf_results_dict, orient="index").reset_index().rename(columns={"index": "MEASUREMENT", 0: "VALUE"}))
                video_clf_pd.insert(loc=0, column="CLASSIFIER", value=clf)
                video_clf_pd.insert(loc=0, column="VIDEO", value=file_name)
                self.results_df = pd.concat([self.results_df, video_clf_pd], axis=0)

    def save(self) -> None:
        """
        Method to save classifier aggregate statistics created in :meth:`~simba.ClfLogCreator.analyze_data` to disk.
        Results are stored in the `project_folder/logs` directory of the SimBA project
        """

        self.results_df = self.results_df[self.results_df["CLASSIFIER"].isin(self.classifiers)].set_index("VIDEO")
        out_df = self.results_df[self.results_df["MEASUREMENT"].isin(self.measurements + [FRAME_COUNT, VIDEO_LENGTH])]
        if not self.transpose:
            self.results_df.to_csv(self.save_path)
        else:
            self.results_df.loc[self.results_df["MEASUREMENT"].isin([FRAME_COUNT, VIDEO_LENGTH]), "CLASSIFIER"] = "Metadata"
            self.results_df = self.results_df.reset_index().drop_duplicates(subset=["VIDEO", "CLASSIFIER", "MEASUREMENT"], keep="first")
            self.results_df["clf_measure"] = (self.results_df["CLASSIFIER"] + " - " + self.results_df["MEASUREMENT"])
            self.results_df = self.results_df.drop(["CLASSIFIER", "MEASUREMENT"], axis=1).reset_index()
            self.results_df = self.results_df.pivot(index="VIDEO", columns="clf_measure", values="VALUE")
            self.results_df.to_csv(self.save_path)

        if self.detailed_bout_data:
            self.bouts_df = pd.concat(self.bouts_df_lst, axis=0)
            self.bouts_df = self.bouts_df[self.bouts_df["Event"].isin(self.classifiers)].set_index("Video")
            if len(self.bouts_df) == 0:
                NoDataFoundWarning(msg=f"No detailed bout data saved: No bouts detected for the selected classifiers: {self.classifiers}")
            else:
                self.bouts_df.to_csv(self.detailed_save_path)
                stdout_success(msg=f"Detailed bout data log saved at {self.detailed_save_path}", source=self.__class__.__name__)
        self.timer.stop_timer()
        stdout_success(msg=f"Data aggregate log saved at {self.save_path}", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)

# test = AggregateClfCalculator(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                               classifiers=['straub_tail'],
#                               transpose=True,
#                               frame_count=True,
#                               video_length=True,
#                               detailed_bout_data=True)
# test.run()
# test.save()




# test = AggregateClfCalculator(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                               data_measures=["Bout count", "Total event duration (s)", "Mean event bout duration (s)", "Median event bout duration (s)", "First event occurrence (s)", "Mean event bout interval duration (s)", "Median event bout interval duration (s)"],
#                               classifiers=['straub_tail'],
#                               video_meta_data = ['Frame count', "Video length (s)"],#
#                               transpose=True)
# test.run()
# test.save()


# test = AggregateClfCalculator(config_path=r"/Users/simon/Desktop/envs/troubleshooting/raph/project_folder/project_config.ini",
#                               data_measures=['Total event duration (s)', 'Median event bout duration (s)'],
#                               classifiers=['walking'],
#                               video_meta_data =['Frame count'],
#                               transpose=True)
#
#
#
# test.run()
# test.save()



from copy import deepcopy
from tkinter import *

from simba.data_processors.movement_calculator import MovementCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, SimbaButton, SimBADropDown)
from simba.utils.checks import check_float
from simba.utils.enums import ConfigKey, Formats, Keys, Links
from simba.utils.errors import NoDataError


class MovementAnalysisPopUp(ConfigReader, PopUpMixin):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if len(self.outlier_corrected_paths) == 0:
            raise NoDataError(msg=f'No data files found in {self.outlier_corrected_dir} directory, cannot compute movement statistics.', source=self.__class__.__name__)
        PopUpMixin.__init__(self, title="ANALYZE MOVEMENT", size=(400, 400), icon='run')
        self.animal_cnt_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT NUMBER OF ANIMALS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.animal_cnt_dropdown = SimBADropDown(parent=self.animal_cnt_frm, label="# OF ANIMALS", label_width=30, dropdown_width=20, value=1, dropdown_options=list(range(1, self.animal_cnt + 1)), command=self.create_settings_frm)
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW, padx=10, pady=10)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)

        self.choose_threshold_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT BODY-PART THRESHOLD", icon_name='threshold', icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.probability_entry = Entry_Box(parent=self.choose_threshold_frm, fileDescription='THRESHOLD: ', labelwidth=30, entry_box_width=20, value=0.0, justify='center')
        self.choose_threshold_frm.grid(row=1, column=0, sticky=NW, padx=10, pady=10)
        self.probability_entry.grid(row=0, column=0, sticky=NW)

        self.body_part_options = deepcopy(self.body_parts_lst)
        for i in self.multi_animal_id_list: self.body_part_options.append(f"{i} CENTER OF GRAVITY")
        self.create_settings_frm(animal_cnt=1)
        self.create_run_frm(run_function=self.run, padx=10, pady=10)
        self.main_frm.mainloop()

    def create_settings_frm(self, animal_cnt):
        if hasattr(self, "bp_frm"):
            self.bp_frm.destroy()
            for k, v in self.body_parts_dropdowns.items():
                v.destroy()

        self.bp_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT BODY-PARTS", icon_name='pose', icon_link=Links.DATA_ANALYSIS.value, padx=5, pady=5, relief='solid')
        self.body_parts_dropdowns = {}
        for cnt, i in enumerate(range(int(animal_cnt))):
            self.body_parts_dropdowns[cnt] = SimBADropDown(parent=self.bp_frm, label=f"Animal {cnt+1}", label_width=30, dropdown_width=20, value=self.body_part_options[cnt], dropdown_options=self.body_part_options)
            self.body_parts_dropdowns[cnt].grid(row=cnt, column=0, sticky=NW)
        self.bp_frm.grid(row=2, column=0, sticky=NW, padx=10, pady=10)

    def run(self):
        self.config.set(ConfigKey.PROCESS_MOVEMENT_SETTINGS.value, ConfigKey.ROI_ANIMAL_CNT.value, str(self.animal_cnt_dropdown.getChoices()))
        check_float(name="Probability threshold", value=self.probability_entry.entry_get, min_value=0.00, max_value=1.00)
        self.config.set(ConfigKey.PROCESS_MOVEMENT_SETTINGS.value, ConfigKey.PROBABILITY_THRESHOLD.value, str(self.probability_entry.entry_get))
        body_parts = []
        threshold = float(self.probability_entry.entry_get)
        for cnt, dropdown in self.body_parts_dropdowns.items():
            body_parts.append(dropdown.getChoices())
        movement_processor = MovementCalculator(config_path=self.config_path,
                                                threshold=threshold,
                                                body_parts=body_parts)
        movement_processor.run()
        movement_processor.save()


# MovementAnalysisPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
#MovementAnalysisPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import os
import shutil
from typing import Any, Dict, List, Optional, Union

import cv2
import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_float, check_if_keys_exist_in_dict,
    check_int, check_that_column_exist, check_valid_lst)
from simba.utils.enums import Formats, TagNames
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df

STYLE_PALETTE = 'palette'
STYLE_SHADING = 'shading'
STYLE_BIN_SIZE = 'bin_size'
STYLE_MAX_SCALE = 'max_scale'

STYLE_ATTR = [STYLE_PALETTE, STYLE_SHADING, STYLE_BIN_SIZE, STYLE_MAX_SCALE]

class HeatmapperLocationSingleCore(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the location where animals spend time.
    For improved run-time, see :meth:`simba.heat_mapper_location_mp.HeatMapperLocationMultiprocess` for multiprocess class.

    ..note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.
        For improved run-time of videos, see :meth:`simba.heat_mapper_location_mp.HeatMapperLocationMultiprocess` for multiprocess class.

    .. image:: _static/img/heatmap_location.gif
       :width: 1000
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param str bodypart: The name of the body-part used to infer the location of the animal.
    :param int bin_size: The rectangular size of each heatmap location in millimeters. For example, `50` will divide the video frames into 5 centimeter rectangular spatial bins.
    :param str palette:  Heatmap pallette. Eg. 'jet', 'magma', 'inferno','plasma', 'viridis', 'gnuplot2'
    :param int or 'auto' max_scale: The max value in the heatmap in seconds. E.g., with a value of `10`, if the classified behavior has occured >= 10s within a rectangular bins, it will be filled with the same color.
    :param bool final_img_setting: If True, create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.

    :example:
    >>> style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'}
    >>> heatmapper = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', style_attr = style_attr, final_img_setting=True, video_setting=True, frame_setting=False, bodypart='Nose_1', files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
    >>> heatmapper.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_paths: List[Union[str, os.PathLike]],
                 bodypart: str,
                 style_attr: Dict[str, Any],
                 final_img_setting: Optional[bool] = True,
                 video_setting: Optional[bool] = False,
                 frame_setting: Optional[bool] = False):


        log_event(logger_name=str(__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(msg="Please choose to select either heatmap videos, frames, and/or final image.", source=self.__class__.__name__)

        check_file_exist_and_readable(file_path=config_path)
        check_valid_lst(data=data_paths, valid_dtypes=(str,), min_len=1)
        check_if_keys_exist_in_dict(data=style_attr, key=STYLE_ATTR, name=f'{self.__class__.__name__} style_attr')
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        self.style_attr = style_attr
        if not os.path.exists(self.heatmap_location_dir):
            os.makedirs(self.heatmap_location_dir)
        self.data_paths = data_paths
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.bp_lst = [f"{self.bp}_x", f"{self.bp}_y"]
        print(f"Processing heatmaps for {len(self.data_paths)} video(s)...")

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.data_paths)
        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(video_name=self.video_name)
            self.width, self.height = int(self.video_info["Resolution_width"].values[0]), int(self.video_info["Resolution_height"].values[0])
            if self.video_setting:
                self.video_save_path = os.path.join(self.heatmap_location_dir, f"{self.video_name}.mp4")
                self.writer = cv2.VideoWriter(self.video_save_path, self.fourcc, self.fps, (self.width, self.height),)
            if self.frame_setting | self.final_img_setting:
                self.save_video_folder = os.path.join(self.heatmap_location_dir, self.video_name)
                if os.path.exists(self.save_video_folder):
                    shutil.rmtree(self.save_video_folder)
                os.makedirs(self.save_video_folder)
            self.data_df = read_df(file_path, self.file_type)
            check_that_column_exist(df=self.data_df, column_name=self.bp_lst, file_name=file_path)
            self.data_df = self.data_df[self.bp_lst]
            squares, aspect_ratio = GeometryMixin().bucket_img_into_grid_square(bucket_grid_size_mm=self.style_attr[STYLE_BIN_SIZE], img_size=(self.width, self.height), px_per_mm=self.px_per_mm)
            location_array = GeometryMixin().cumsum_coord_geometries(data=self.data_df.values, fps=self.fps, geometries=squares)

            if self.style_attr[STYLE_MAX_SCALE] == "auto":
                self.max_scale = np.round(np.max(np.max(location_array[-1], axis=0)), 3)
                if self.max_scale == 0: self.max_scale = 1
            else:
                check_float(name=f'{self.__class__.__name__} style max scale', value=self.style_attr[STYLE_MAX_SCALE], min_value=10e-6)
                self.max_scale = self.style_attr[STYLE_MAX_SCALE]

            if self.final_img_setting:
                final_img_save_name = os.path.join(self.heatmap_location_dir, f"{self.video_name}_final_frm.png")
                self.make_location_heatmap_plot(frm_data=location_array[-1:, :, :][0],
                                                max_scale=self.max_scale,
                                                palette=self.style_attr[STYLE_PALETTE],
                                                aspect_ratio=aspect_ratio,
                                                file_name=final_img_save_name,
                                                shading=self.style_attr[STYLE_SHADING],
                                                img_size=(self.width, self.height))
                print(f"Final heatmap image saved at {final_img_save_name}")

            if (self.frame_setting) or (self.video_setting):
                for frm_cnt, cumulative_frm in enumerate(range(location_array.shape[0])):
                    img = self.make_location_heatmap_plot(frm_data=location_array[cumulative_frm, :, :],
                                                          max_scale=self.max_scale,
                                                          palette=self.style_attr["palette"],
                                                          aspect_ratio=aspect_ratio,
                                                          file_name=None,
                                                          shading=self.style_attr["shading"],
                                                          img_size=(self.width, self.height))
                    if self.video_setting:
                        self.writer.write(img[:, :, :3])
                    if self.frame_setting:
                        frame_save_path = os.path.join(self.save_video_folder, f"{frm_cnt}.png")
                        cv2.imwrite(frame_save_path, img)

                    print(f"Heatmap frame: {frm_cnt + 1} / {len(self.data_df)}. Video: {self.video_name} ({file_cnt + 1}/{len(self.data_paths)})")

            if self.video_setting:
                self.writer.release()
            video_timer.stop_timer()
            print(f"Heatmap plot for video {self.video_name} saved (elapsed time: {video_timer.elapsed_time_str}s")
        self.timer.stop_timer()
        stdout_success(msg=f"Created heatmaps for {len(self.data_paths)} videos", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__,)


# test = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/platea/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'flat', 'bin_size': 30, 'max_scale': 2},
#                                       final_img_setting=True,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='CENTER',
#                                       data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/platea/project_folder/csv/outlier_corrected_movement_location/Video_1.csv'])
# test.run()






# test = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/open_field_below/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 10},
#                                       final_img_setting=True,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='Snout',
#                                       data_paths=['/Users/simon/Desktop/envs/simba/troubleshooting/open_field_below/project_folder/csv/outlier_corrected_movement_location/raw_clip1.csv'])
# test.run()




#
# test = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='Nose_1',
#                                       files_found=['/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement_location/Together_1.csv'])
# test.run()


# test = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 50, 'max_scale': 'auto'},
#                                       final_img_setting=False,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='Nose',
#                                       files_found=['/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/csv/outlier_corrected_movement_location/PD1406_2022-05-24_RVDG_GCaMP8s-Gi_Video_Day_22_Baseline.csv'])
# test.create_heatmaps()


# test = HeatmapperLocation(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini',
#                           bodypart='Nose_1',
#                           bin_size=50,
#                           palette='jet',
#                           max_scale='auto',
#                           final_img_setting=True,
#                           video_setting=False,
#                           frame_setting=False)
# test.create_heatmaps()



__author__ = "Simon Nilsson"

import os
import threading
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.probability_plot_creator import \
    TresholdPlotCreatorSingleProcess
from simba.plotting.probability_plot_creator_mp import \
    TresholdPlotCreatorMultiprocess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, SimbaButton,
                                        SimbaCheckbox, SimBADropDown)
from simba.utils.enums import Formats, Links
from simba.utils.lookups import get_color_dict
from simba.utils.read_write import (check_if_filepath_list_is_empty,
                                    get_file_name_info_in_directory)

STYLE_WIDTH = 'width'
STYLE_HEIGHT = 'height'
STYLE_FONT_SIZE = 'font size'
STYLE_LINE_WIDTH = 'line width'
STYLE_YMAX = 'y_max'
STYLE_COLOR = 'color'
STYLE_OPACITY = 'opacity'
AUTO = 'AUTO'

MAX_Y_OPTIONS = list(range(100, 0, -10))
MAX_Y_OPTIONS.insert(0, AUTO)
OPACITY_OPTIONS = [round(x * 0.1, 1) for x in range(1, 11)]

class VisualizeClassificationProbabilityPopUp(PopUpMixin, ConfigReader):

    """
    :example:
    >>> _ = VisualizeClassificationProbabilityPopUp(config_path=r'C:\troubleshooting\RAT_NOR\project_folder\project_config.ini')
    """


    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        color_names = list(get_color_dict().keys())

        self.files_found_dict = get_file_name_info_in_directory(directory=self.machine_results_dir, file_type=self.file_type)
        check_if_filepath_list_is_empty(filepaths=list(self.files_found_dict.keys()), error_msg=f"Cannot visualize probabilities, no data in {self.machine_results_dir} directory")
        max_file_name_len = max(len(k) for k in self.files_found_dict) + 5
        PopUpMixin.__init__(self, title="CREATE CLASSIFICATION PROBABILITY PLOTS", icon='probability')

        self.style_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="STYLE SETTINGS", icon_name='style', icon_link=Links.VISUALIZE_CLF_PROBABILITIES.value, pady=5, padx=5, relief='solid')
        self.resolution_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=self.resolutions, label='RESOLUTION: ', label_width=25, dropdown_width=35, value=self.resolutions[1])
        self.max_y_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=MAX_Y_OPTIONS, label='MAX Y-AXIS: ', label_width=25, dropdown_width=35, value=AUTO)
        self.line_clr_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=color_names, label='LINE COLOR: ', label_width=25, dropdown_width=35, value='Red')
        self.font_size_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=list(range(1, 26)), label='TEXT SIZE: ', label_width=25, dropdown_width=35, value=10)
        self.line_width_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=list(range(1, 26)), label='LINE WIDTH: ', label_width=25, dropdown_width=35, value=6)
        self.line_opacity_dropdown = SimBADropDown(parent=self.style_settings_frm, dropdown_options=OPACITY_OPTIONS, label='LINE OPACITY: ', label_width=25, dropdown_width=35, value=1.0)


        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="VISUALIZATION SETTINGS", icon_name='eye', icon_link=Links.VISUALIZE_CLF_PROBABILITIES.value, pady=5, padx=5, relief='solid')
        self.clf_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=self.clf_names, label='CLASSIFIER: ', label_width=25, dropdown_width=35, value=self.clf_names[0])
        self.core_cnt_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=list(range(1, self.cpu_cnt+1)), label='CPU CORE COUNT: ', label_width=25, dropdown_width=35, value=int(self.cpu_cnt/2))

        probability_frames_cb, self.probability_frames_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE FRAMES', font=Formats.FONT_REGULAR.value, txt_img='frames', val=False)
        probability_videos_cb, self.probability_videos_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE VIDEOS', font=Formats.FONT_REGULAR.value, txt_img='video', val=False)
        probability_last_frm_cb, self.probability_last_frm_var = SimbaCheckbox(parent=self.settings_frm, txt='CREATE FINAL FRAME', font=Formats.FONT_REGULAR.value, txt_img='finish', val=True)


        self.run_single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SINGLE VIDEO", icon_name='video', icon_link=Links.VISUALIZE_CLF_PROBABILITIES.value, pady=5, padx=5, relief='solid')
        self.run_single_video_btn = SimbaButton(parent=self.run_single_video_frm, txt="CREATE SINGLE VIDEO", img='rocket', txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.__run, cmd_kwargs={'multiple': False})
        self.single_video_dropdown = SimBADropDown(parent=self.run_single_video_frm, dropdown_options=list(self.files_found_dict.keys()), label='VIDEO: ', label_width=25, dropdown_width=max_file_name_len, value=list(self.files_found_dict.keys())[0])

        self.run_multiple_videos = CreateLabelFrameWithIcon(parent=self.main_frm, header="MULTIPLE VIDEO", icon_name='video', icon_link=Links.VISUALIZE_CLF_PROBABILITIES.value, pady=5, padx=5, relief='solid')
        self.run_multiple_video_btn = SimbaButton(parent=self.run_multiple_videos, txt=f"Create multiple videos ({len(list(self.files_found_dict.keys()))} video(s) found)", img='rocket', txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.__run, cmd_kwargs={'multiple': True})

        self.style_settings_frm.grid(row=0, sticky=NW, padx=10, pady=10)
        self.resolution_dropdown.grid(row=0, sticky=NW)
        self.line_clr_dropdown.grid(row=1, sticky=NW)
        self.font_size_dropdown.grid(row=2, sticky=NW)
        self.line_width_dropdown.grid(row=3, sticky=NW)
        self.max_y_dropdown.grid(row=4, sticky=NW)
        self.line_opacity_dropdown.grid(row=5, sticky=NW)

        self.settings_frm.grid(row=1, sticky=NW, padx=10, pady=10)
        self.clf_dropdown.grid(row=0, sticky=NW)
        self.core_cnt_dropdown.grid(row=1, sticky=NW)
        probability_frames_cb.grid(row=2, sticky=NW)
        probability_videos_cb.grid(row=3, sticky=NW)
        probability_last_frm_cb.grid(row=4, sticky=NW)


        self.run_single_video_frm.grid(row=2, sticky=NW, padx=10, pady=10)
        self.run_single_video_btn.grid(row=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos.grid(row=3, sticky=NW, padx=10, pady=10)
        self.run_multiple_video_btn.grid(row=0, sticky=NW)

        self.main_frm.mainloop()

    def __run(self, multiple: bool):
        width = int(self.resolution_dropdown.getChoices().split("×")[0])
        height = int(self.resolution_dropdown.getChoices().split("×")[1])
        font_size = int(self.font_size_dropdown.get_value())
        line_width = int(self.line_width_dropdown.get_value())
        core_cnt = int(self.core_cnt_dropdown.get_value())
        opacity = float(self.line_opacity_dropdown.get_value())


        style_attr = {
            STYLE_WIDTH: width,
            STYLE_HEIGHT: height,
            STYLE_FONT_SIZE: font_size,
            STYLE_LINE_WIDTH: line_width,
            STYLE_COLOR: self.line_clr_dropdown.getChoices(),
            STYLE_YMAX: self.max_y_dropdown.getChoices(),
            STYLE_OPACITY: opacity}

        if multiple:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [self.files_found_dict[self.single_video_dropdown.getChoices()]]

        if core_cnt == 1:
            probability_plot_creator = TresholdPlotCreatorSingleProcess(config_path=self.config_path,
                                                                        frame_setting=self.probability_frames_var.get(),
                                                                        video_setting=self.probability_videos_var.get(),
                                                                        last_image=self.probability_last_frm_var.get(),
                                                                        files_found=data_paths,
                                                                        clf_name=self.clf_dropdown.getChoices(),
                                                                        style_attr=style_attr)
        else:
            probability_plot_creator = TresholdPlotCreatorMultiprocess(config_path=self.config_path,
                                                                       frame_setting=self.probability_frames_var.get(),
                                                                       video_setting=self.probability_videos_var.get(),
                                                                       last_frame=self.probability_last_frm_var.get(),
                                                                       files_found=data_paths,
                                                                       clf_name=self.clf_dropdown.getChoices(),
                                                                       cores=core_cnt,
                                                                       style_attr=style_attr)

        probability_plot_creator.run()

        #threading.Thread(target=probability_plot_creator.run).start()




#_ = VisualizeClassificationProbabilityPopUp(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini")
#_ = VisualizeClassificationProbabilityPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')



import os
from typing import Optional, Tuple, Union

import numpy as np

from simba.mixins.geometry_mixin import GeometryMixin
from simba.plotting.geometry_plotter import GeometryPlotter
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_if_valid_rgb_tuple,
                                check_int, check_valid_dataframe,
                                check_video_and_data_frm_count_align)
from simba.utils.enums import Formats
from simba.utils.errors import NoFilesFoundError
from simba.utils.lookups import get_current_time
from simba.utils.read_write import (find_all_videos_in_directory,
                                    find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_df)

VERTICE = 'vertice'
CENTER_X = 'center_x'
CENTER_Y = 'center_y'
NOSE_X = 'nose_x'
NOSE_Y = 'nose_y'
TAIL_X = 'tail_x'
TAIL_Y = 'tail_y'
LEFT_X = 'left_x'
LEFT_Y = 'left_y'
RIGHT_X = 'right_x'
RIGHT_Y = 'right_y'

BP_COLS = [NOSE_X, NOSE_Y, TAIL_X, TAIL_Y, LEFT_X, LEFT_Y, RIGHT_X, RIGHT_Y]

class BlobVisualizer():
    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 core_cnt: int = -1,
                 shape_opacity: float = 0.5,
                 bg_opacity: float = 1.0,
                 circle_size: Optional[int] = None,
                 hull: Optional[Tuple[int, int, int]] = (178, 102, 255),
                 anterior: Optional[Tuple[int, int, int]] = (0, 0, 255),
                 posterior: Optional[Tuple[int, int, int]] = (0, 128, 0),
                 center: Optional[Tuple[int, int, int]] = (0, 165, 255),
                 left: Optional[Tuple[int, int, int]] = (255, 51, 153),
                 right: Optional[Tuple[int, int, int]] = (255, 255, 102)):

        if os.path.isdir(data_path):
            self.data_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=['.csv'], raise_error=True, as_dict=True)
        elif os.path.isfile(data_path):
            check_file_exist_and_readable(file_path=data_path)
            self.data_paths = {get_fn_ext(filepath=data_path)[1]: data_path}
        else:
            raise NoFilesFoundError(msg=f'{data_path} is not a valid file path or directory.', source=self.__class__.__name__)
        if os.path.isdir(video_path):
            self.video_paths = find_all_videos_in_directory(directory=video_path, raise_error=True, as_dict=True)
        elif os.path.isfile(video_path):
            check_file_exist_and_readable(file_path=video_path)
            self.video_paths = {get_fn_ext(filepath=video_path)[1]: video_path}
        else:
            raise NoFilesFoundError(msg=f'{video_path} is not a valid directory of file path.')
        for k, v in self.data_paths.items():
            if k not in self.video_paths.keys():
                raise NoFilesFoundError(msg=f'Could not find a video for blob tracking data file {k} in the {video_path} directory', source=self.__class__.__name__)
        check_if_dir_exists(in_dir=save_dir, create_if_not_exist=True)
        check_int(name='core count', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        self.core_cnt = find_core_cnt()[0] if core_cnt == -1 or core_cnt > find_core_cnt()[0] else core_cnt
        check_float(name='shape_opacity', value=shape_opacity, max_value=1.0, min_value=0.1, raise_error=True)
        check_float(name='bg_opacity', value=bg_opacity, max_value=1.0, min_value=0.1, raise_error=True)
        if circle_size is not None:
            check_int(name='circle_size', value=circle_size, min_value=1)
        self.circle_size = circle_size
        self.bg_opacity, self.shape_opacity = bg_opacity, shape_opacity
        self.save_dir = save_dir
        for i in [hull, anterior, posterior, center, left, right]:
            if i is not None:
                check_if_valid_rgb_tuple(data=i)
        self.anterior, self.posterior, self.center, self.left, self.right, self.hull = anterior, posterior, center, left, right, hull
        self.colors = [x for x in [self.hull, self.anterior, self.posterior, self.center, self.left, self.right] if x is not None]

    def run(self):
        print(f'Running blob visualization for {len(list(self.data_paths.keys()))} video(s) (start time: {get_current_time()})...')
        for file_name, file_path in self.data_paths.items():
            print(f'Running blob visualization for video {file_name}...')
            geometries = []
            video_path = self.video_paths[file_name]
            df = read_df(file_path=file_path, file_type='csv')
            check_video_and_data_frm_count_align(video=video_path, data=df, name=file_name, raise_error=True)
            check_valid_dataframe(df=df, source='', valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=BP_COLS)
            if self.hull:
                vertice_cols = [x for x in df.columns if VERTICE in x]
                vertice_arr = df[vertice_cols].values.reshape(len(df), int(len(vertice_cols) / 2), 2).astype(np.int32)
                geometries.append(GeometryMixin().bodyparts_to_polygon(data=vertice_arr, simplify_tolerance=2, convex_hull=False))
            if self.anterior is not None:
                geometries.append(GeometryMixin.bodyparts_to_points(data=df[[NOSE_X, NOSE_Y]].values))
            if self.posterior is not None:
                geometries.append(GeometryMixin.bodyparts_to_points(data=df[[TAIL_X, TAIL_Y]].values))
            if self.center is not None:
                geometries.append(GeometryMixin.bodyparts_to_points(data=df[[CENTER_X, CENTER_Y]].values))
            if self.left is not None:
                geometries.append(GeometryMixin.bodyparts_to_points(data=df[[LEFT_X, LEFT_Y]].values))
            if self.right is not None:
                geometries.append(GeometryMixin.bodyparts_to_points(data=df[[RIGHT_X, RIGHT_Y]].values))

            plotter = GeometryPlotter(geometries=geometries,
                                      video_name=video_path,
                                      save_dir=self.save_dir,
                                      core_cnt=self.core_cnt,
                                      palette=None,
                                      colors=self.colors,
                                      bg_opacity=self.bg_opacity,
                                      shape_opacity=self.shape_opacity,
                                      circle_size=self.circle_size)
            plotter.run()
            # #

            # img = read_frm_of_video(video_path=video_path, frame_index=0)
            #
            # cv2.imshow('asdasdas', img)
            # cv2.waitKey(5000)



# visualizer = BlobVisualizer(data_path=r'D:\troubleshooting\netholabs\original_videos\whiskers\results',
#                             video_path=r'D:\troubleshooting\netholabs\original_videos\whiskers',
#                             save_dir=r"D:\troubleshooting\netholabs\original_videos\whiskers\out_video",
#                             core_cnt=15,
#                             posterior=None,
#                             left=None,
#                             right=None)
# visualizer.run()

# visualizer = BlobVisualizer(data_path=r'D:\EPM_4\data', video_path=r'D:\EPM_4\original', save_dir=r"D:\EPM_4\out_videos")
# visualizer.run()


# visualizer = BlobVisualizer(data_path=r'D:\water_maze\data', video_path=r'D:\water_maze\original', save_dir=r"D:\water_maze\out_videos")
# visualizer.run()
#
# #
#visualizer = BlobVisualizer(data_path=r'D:\open_field_3\sample\blob_data', video_path=r'D:\open_field_3\sample', save_dir=r"D:\open_field_3\sample\out_videos")
#visualizer = BlobVisualizer(data_path=r'D:\EPM\large_sample\out', video_path=r'D:\EPM\large_sample', save_dir=r'D:\EPM\videos_out')

#visualizer = BlobVisualizer(data_path=r'D:\EPM\sampled\data', video_path=r'D:\EPM\sampled', save_dir=r"D:\EPM\sampled\out_videos")
# visualizer = BlobVisualizer(data_path=r'D:\open_field_2\sample\clipped_10min\data', video_path=r'D:\open_field_2\sample\clipped_10min', save_dir=r"D:\open_field_2\sample\clipped_10min\videos_out")

#visualizer = BlobVisualizer(data_path=r'D:\open_field_4\data', video_path=r'D:\open_field_4', save_dir="D:\open_field_4\out_videos")
#visualizer = BlobVisualizer(data_path=r'D:\open_field\data', video_path=r'D:\open_field', save_dir=r"D:\open_field\videos")
#visualizer = BlobVisualizer(data_path=r'D:\EPM_3\out', video_path=r'D:\EPM_3', save_dir=r"D:\EPM_3\out_videos")

# visualizer = BlobVisualizer(data_path=r'D:\FST_SIDE\out', video_path=r'D:\FST_SIDE', save_dir=r"D:\FST_SIDE\videos")
# #
# # #visualizer = BlobVisualizer(data_path=r'D:\EPM\data', video_path=r'D:\EPM\sampled', save_dir=r"D:\EPM\videos_out")
# visualizer.run()

# visualizer = BlobVisualizer(data_path=r'D:\OF_7\data', video_path=r'D:\OF_7\original', save_dir=r'D:\OF_7\out_videos')
# visualizer.run()


__author__ = "Simon Nilsson"

import multiprocessing
import os
from tkinter import *
from typing import Union

from simba.data_processors.timebins_clf_calculator import TimeBinsClfCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaButton, SimbaCheckbox)
from simba.utils.checks import check_int
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.errors import (NoChoosenClassifierError,
                                NoChoosenMeasurementError, NoDataError)


class TimeBinsClfPopUp(PopUpMixin, ConfigReader):
    def __init__(self,
                 config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if len(self.machine_results_paths) == 0:
            raise NoDataError(msg=f'Cannot compute classifications by time-bin: no data found in the {self.machine_results_dir} directory', source=self.__class__.__name__)
        PopUpMixin.__init__(self, title="CLASSIFICATION BY TIME BINS", icon='timer_2')

        measures_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="MEASUREMENTS", icon_name='ruler', icon_link=Links.ANALYZE_ML_RESULTS.value, padx=5, pady=5, relief='solid')
        first_occurance_cb, self.first_occurance_var = SimbaCheckbox(parent=measures_frm, txt="First occurrence (s)", txt_img='one', val=True)
        event_count_cb, self.event_count_var = SimbaCheckbox(parent=measures_frm, txt="First occurrence (s)", txt_img='abacus', val=True)
        total_event_duration_cb, self.total_event_duration_var = SimbaCheckbox(parent=measures_frm, txt="Total event duration (s)", txt_img='stopwatch', val=True)
        mean_event_duration_cb, self.mean_event_duration_var = SimbaCheckbox(parent=measures_frm, txt="Mean event duration (s)", txt_img='average', val=False)
        median_event_duration_cb, self.median_event_duration_var = SimbaCheckbox(parent=measures_frm, txt="Median event duration (s)", txt_img='average_2', val=False)
        mean_interval_duration_cb, self.mean_interval_duration_var = SimbaCheckbox(parent=measures_frm, txt="Mean event interval (s)", txt_img='timer_2', val=False)
        median_interval_duration_cb, self.median_interval_duration_var = SimbaCheckbox(parent=measures_frm, txt="Median event interval (s)", txt_img='timer', val=False)

        measures_frm.grid(row=0, column=0, sticky=NW)
        first_occurance_cb.grid(row=0, column=0, sticky=NW)
        event_count_cb.grid(row=1, column=0, sticky=NW)
        total_event_duration_cb.grid(row=2, column=0, sticky=NW)
        mean_event_duration_cb.grid(row=3, column=0, sticky=NW)
        median_event_duration_cb.grid(row=4, column=0, sticky=NW)
        mean_interval_duration_cb.grid(row=5, column=0, sticky=NW)
        median_interval_duration_cb.grid(row=6, column=0, sticky=NW)

        self.clf_vars = {}
        clf_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="CLASSIFIERS", icon_name='forest', icon_link=Links.ANALYZE_ML_RESULTS.value, padx=5, pady=5, relief='solid')
        for cnt, clf_name in enumerate(self.clf_names):
            cb, self.clf_vars[clf_name] = SimbaCheckbox(parent=clf_frm, txt=clf_name, val=True)
            cb.grid(row=cnt, column=0, sticky=NW)

        clf_frm.grid(row=1, column=0, sticky=NW, padx=10, pady=10)
        time_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SET TIME", icon_name='timer_2', icon_link=Links.ANALYZE_ML_RESULTS.value, padx=5, pady=5,  relief='solid')
        self.timebin_entrybox = Entry_Box(time_frm, "TIME BIN SIZE (S): ", validation="numeric", labelwidth=30, entry_box_width=15, value=30)

        time_frm.grid(row=2, column=0, sticky=NW)
        self.timebin_entrybox.grid(row=0, column=0, sticky=NW, pady=10, padx=10)

        run_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="RUN", icon_name='run', icon_link=Links.ANALYZE_ML_RESULTS.value, padx=5, pady=5, relief='solid')
        run_button = SimbaButton(parent=run_frm, txt="Run", img='rocket', font=Formats.FONT_REGULAR.value, cmd=self._run)

        run_frm.grid(row=3, column=0, sticky=NW)
        run_button.grid(row=0, column=0, sticky=NW, pady=10)

        self.main_frm.mainloop()

    def _run(self):
        check_int(name="Time bin", value=self.timebin_entrybox.entry_get)

        first_occurance = self.first_occurance_var.get()
        event_cnt = self.event_count_var.get()
        total_event_duration = self.total_event_duration_var.get()
        mean_event_duration = self.mean_event_duration_var.get()
        median_event_duration = self.median_event_duration_var.get()
        mean_interval_duration = self.mean_interval_duration_var.get()
        median_interval_duration = self.median_interval_duration_var.get()

        if not any([first_occurance, event_cnt, total_event_duration, mean_event_duration, median_event_duration, mean_interval_duration, median_interval_duration]):
            raise NoChoosenMeasurementError(source=self.__class__.__name__)

        clfs = [clf_name for clf_name, cb_var in self.clf_vars.items() if cb_var.get()]
        if len(clfs) == 0:
            raise NoChoosenClassifierError(source=self.__class__.__name__)

        time_bins_clf_analyzer = TimeBinsClfCalculator(config_path=self.config_path,
                                                       bin_length=int(self.timebin_entrybox.entry_get),
                                                       classifiers=clfs,
                                                       first_occurrence=first_occurance,
                                                       event_count=event_cnt,
                                                       total_event_duration=total_event_duration,
                                                       mean_event_duration=mean_event_duration,
                                                       median_event_duration=median_event_duration,
                                                       mean_interval_duration=mean_interval_duration,
                                                       median_interval_duration=median_interval_duration)

        time_bins_clf_analyzer.run()

#_ = TimeBinsClfPopUp(config_path=r"D:\troubleshooting\mitra\project_folder\project_config.ini")

#_ = TimeBinsClfPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')



import os

os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
try:
    from typing import Literal
except:
    from typing_extensions import Literal

from typing import List, Optional, Union

import numpy as np
import pandas as pd
import torch

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int,
                                check_valid_boolean, check_valid_lst,
                                get_fn_ext)
from simba.utils.data import resample_geometry_vertices
from simba.utils.enums import Options
from simba.utils.errors import InvalidFileTypeError
from simba.utils.printing import SimbaTimer
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_video_meta_data)
from simba.utils.yolo import load_yolo_model, yolo_predict

TASK = 'segment'

class YOLOSegmentationInference():

    """
     Run inference on video(s) using a trained YOLO segmentation model.

    :param Union[str, os.Pathlike] weights_path: Path to the trained YOLO `.pt` weights file.
    :param Union[str, os.Pathlike]  video_path: Path to a single video or a list of video paths to run inference on.
    :param bool verbose: Whether to print progress information. Default is True.
    :param Union[str, os.Pathlike]  save_dir: Directory where output videos and data will be saved.
    :param Union[str, int] device: Device to run inference on; use 'cpu' or an integer GPU index (e.g., 0).
    :param str format: Optional export format for the model. Supported values: "onnx", "engine", "torchscript", "onnxsimplify", "coreml", "openvino", "pb", "tf", "tflite". Defaults to None.
    :param Optional[int] batch_size: Number of frames to process at once. Increase for faster performance with sufficient memory.
    :param int torch_threads: Number of CPU threads to use (when on CPU).
    :param bool half_precision: Whether to use half-precision (FP16) for inference on GPU. Default is True.
    :param bool stream: Whether to stream video processing (less memory, suitable for long videos).
    :param float threshold: Confidence threshold for object/segmentation detection.
    :param int max_tracks: Optional maximum number of objects to track. If None, tracking is disabled.
    :param bool interpolate: Whether to interpolate results (useful for smoothing or low-FPS videos).
    :param int imgsz: Inference image size (width/height in pixels); must be multiple of 32.
    :param float iou: IoU threshold for non-max suppression (NMS).
    :param bool retina_msk: Whether to use high-resolution Retina-style masks.
    :param int vertice_cnt: Number of vertices used to approximate the segmentation mask polygon.

    .. note::
       To **create** YOLO segmentation dataset for fitting, use :func:`simba.third_party_label_appenders.transform.labelme_to_yolo_seg.LabelmeKeypoints2YoloSeg`.
       To fit YOLO model, see `:func:~simba.bounding_box_tools.yolo.yolo_fit.FitYolo`.
       To visualize the segmentation results, see :func:`simba.bounding_box_tools.yolo.yolo_seg_visualizer.YOLOSegmentationVisualizer`

    :example:
    >>> weights_path = r"D:\platea\yolo_071525\mdl\train3\weights\best.pt"
    >>> video_path = r"D:\platea\platea_videos\videos\clipped\10B_Mouse_5-choice_MustTouchTrainingNEWFINAL_a7.mp4"
    >>> save_dir=r"D:\platea\platea_videos\videos\yolo_results"
    >>> runner = YOLOSegmentationInference(weights_path=weights_path, video_path=video_path, save_dir=save_dir, verbose=True, device=0, format=None, stream=True, batch_size=10, imgsz=320, interpolate=True, threshold=0.8, retina_msk=True)
    >>> runner.run()

    """

    def __init__(self,
                 weights_path: Union[str, os.PathLike],
                 video_path: Union[Union[str, os.PathLike], List[Union[str, os.PathLike]]],
                 verbose: Optional[bool] = True,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 device: Union[Literal['cpu'], int] = 0,
                 format: Optional[str] = None,
                 batch_size: Optional[int] = 4,
                 torch_threads: int = 8,
                 half_precision: bool = True,
                 stream: bool = False,
                 threshold: float = 0.5,
                 max_tracks: Optional[int] = None,
                 interpolate: bool = False,
                 imgsz: int = 640,
                 iou: float = 0.5,
                 retina_msk: Optional[bool] = False,
                 vertice_cnt: int = 30):

        if isinstance(video_path, list):
            check_valid_lst(data=video_path, source=f'{self.__class__.__name__} video_path', valid_dtypes=(str, np.str_,), min_len=1)
        elif os.path.isfile(video_path):
            video_path = [video_path]
        elif os.path.isdir(video_path):
            video_path = find_files_of_filetypes_in_directory(directory=video_path, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, as_dict=False)
        for i in video_path:
            _ = get_video_meta_data(video_path=i)
        check_file_exist_and_readable(file_path=weights_path)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose')
        check_valid_boolean(value=interpolate, source=f'{self.__class__.__name__} interpolate')
        check_valid_boolean(value=retina_msk, source=f'{self.__class__.__name__} retina_msk')
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1)
        check_int(name=f'{self.__class__.__name__} imgsz', value=imgsz, min_value=1)
        check_int(name=f'{self.__class__.__name__} vertice_cnt', value=vertice_cnt, min_value=3)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=10e-6, max_value=1.0)
        check_float(name=f'{self.__class__.__name__} iou', value=iou, min_value=10e-6, max_value=1.0)
        if max_tracks is not None:
            check_int(name=f'{self.__class__.__name__} max_tracks', value=max_tracks, min_value=1)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir')
        torch.set_num_threads(torch_threads)
        self.model = load_yolo_model(weights_path=weights_path, device=device, format=format)
        self.half_precision, self.stream, self.video_path, self.retina_msk = half_precision, stream, video_path, retina_msk
        self.device, self.batch_size, self.threshold, self.max_tracks, self.iou = device, batch_size, threshold, max_tracks, iou
        self.verbose, self.save_dir, self.imgsz, self.interpolate = verbose, save_dir, imgsz, interpolate
        self.vertice_cnt = vertice_cnt
        if self.model.model.task != TASK:
            raise InvalidFileTypeError(msg=f'The model {weights_path} is not a segmentation model. It is a {self.model.model.task} model', source=self.__class__.__name__)
        self.vertice_col_names = ['FRAME', 'ID']
        for i in range(self.vertice_cnt):
            self.vertice_col_names.append(f"VERTICE_{i}_X"); self.vertice_col_names.append(f"VERTICE_{i}_Y")
    def run(self):
        results = {}
        class_dict = self.model.names
        timer = SimbaTimer(start=True)
        for path in self.video_path:
            _, video_name, _ = get_fn_ext(filepath=path)
            video_meta_data = get_video_meta_data(video_path=path)
            video_results = []
            video_predictions = yolo_predict(model=self.model,
                                             source=path,
                                             half=self.half_precision,
                                             batch_size=self.batch_size,
                                             stream=self.stream,
                                             imgsz=self.imgsz,
                                             device=self.device,
                                             threshold=self.threshold,
                                             max_detections=self.max_tracks,
                                             verbose=self.verbose,
                                             iou=self.iou,
                                             retina_msk=self.retina_msk)
            for frm_cnt, video_prediction in enumerate(video_predictions):
                if video_prediction.masks is not None:
                    boxes = video_prediction.obb.data if video_prediction.obb is not None else video_prediction.boxes.data
                    boxes = boxes.cpu().numpy().astype(np.float32)
                    detected_classes = boxes[:, -1].astype(int) if boxes.size > 0 else []
                    detected_masks = video_prediction.masks.xy
                    for detection_cnt, detected_class in enumerate(detected_classes):
                        mask = detected_masks[detection_cnt].reshape(-1, detected_masks[detection_cnt].shape[0], 2)
                        vertices = resample_geometry_vertices(vertices=mask, vertice_cnt=self.vertice_cnt).flatten()
                        vertices = np.insert(vertices, 0, detected_class)
                        vertices = np.insert(vertices, 0, frm_cnt)
                        video_results.append(vertices)

            vertices = pd.DataFrame(video_results, columns=self.vertice_col_names)
            if self.save_dir:
                save_path = os.path.join(self.save_dir, f'{video_name}.csv')
                vertices.to_csv(save_path)
            else:
                results[video_name] = vertices

        timer.stop_timer()
        if not self.save_dir:
            if self.verbose: print(f'YOLO results created', timer.elapsed_time_str)
            return results
        else:
            if self.verbose:
                print(f'YOLO results saved in {self.save_dir} directory', timer.elapsed_time_str)
            return None

# weights_path = r"D:\platea\yolo_071525\mdl\train3\weights\best.pt"
# video_path = r"D:\platea\platea_videos\videos\clipped\10B_Mouse_5-choice_MustTouchTrainingNEWFINAL_a7.mp4"
# save_dir=r"D:\platea\platea_videos\videos\yolo_results"
# i = YOLOSegmentationInference(weights_path=weights_path,
#                              video_path=video_path,
#                              save_dir=save_dir,
#                              verbose=True,
#                              device=0,
#                              format=None,
#                              stream=True,
#                              batch_size=10,
#                              imgsz=320,
#                              interpolate=True,
#                              threshold=0.8,
#                              retina_msk=True)
# i.run()


import os
import random
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.third_party_label_appenders.transform.utils import (
    create_yolo_keypoint_yaml, get_yolo_keypoint_flip_idx)
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int, check_str,
                                check_valid_boolean, check_valid_dataframe,
                                check_valid_tuple,
                                check_video_and_data_frm_count_align)
from simba.utils.enums import Formats, Options
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data, read_df,
                                    read_frm_of_video)
from simba.utils.warnings import NoDataFoundWarning
from simba.utils.yolo import keypoint_array_to_yolo_annotation_str


class SimBA2Yolo:
    """
    Convert pose estimation data from a SimBA project into the YOLO keypoint format, including frame sampling,
    image-label pair creation, bounding box computation, and train/validation splitting.

    .. note::
       For creating the ``flip_idx``, see :func:`simba.third_party_label_appenders.converters.get_yolo_keypoint_flip_idx`.
       For creating the ``bp_id_idx``, see :func:`simba.third_party_label_appenders.converters.get_yolo_keypoint_bp_id_idx`

    :param Union[str, os.PathLike] config_path: Path to the SimBA project `.ini` configuration file.
    :param Union[str, os.PathLike] save_dir: Directory where YOLO-formatted data will be saved. Subdirectories for images/labels (train/val) are created.
    :param Optional[Union[str, os.PathLike] data_dir: Optional directory containing outlier-corrected SimBA pose estimation data. If None, uses path from config.
    :param float train_size: Proportion of samples to allocate to the training set (range 0.1–0.99). Remaining samples go to validation.
    :param bool verbose: If True, prints progress updates to the console.
    :param bool greyscale: If True, saves extracted video frames in greyscale. Otherwise, saves in color.
    :param float padding: Padding added around the bounding box (as a proportion of image dimensions, range 0.0–1.0). Useful if animal body-parts are in a "line".
    :param Tuple[int, ...] flip_idx: Tuple defining symmetric keypoint indices for horizontal flipping. Used to write the `map.yaml` file. If None, then attempt to infer.
    :param Dict[int, str] names: Dictionary mapping instance IDs to class names. Used in annotation labels and `map.yaml`.
    :param Optional[int] sample_size: If specified, limits the number of randomly sampled frames per video. If None, all frames are used.
    :param Optional[Dict[int, Union[Tuple[int], List[int]]]] bp_id_idx: Optional mapping of instance IDs to keypoint index groups, allowing support for multiple animals per frame. Must match keys in `map_dict`.
    :param Optional[str] single_id: If the data contains pose-estimation for multiple indivisuals, but you want to treat it as examples of a single individual, pass the name of the single individual. Defaults to None, and the YOLO data will be formatted to the number of objects which the H5 data contains.
    :return: None. Saves YOLO-formatted images and annotations to disk in the `save_dir` location.

    :example:
    >>> SAVE_DIR = r'D:\troubleshooting\mitra\mitra_yolo'
    >>> CONFIG_PATH = r"C:\troubleshooting\mitra\project_folder\project_config.ini"
    >>> runner = SimBA2Yolo(config_path=CONFIG_PATH, save_dir=SAVE_DIR, sample_size=10, verbose=True)
    >>> runner.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 train_size: float = 0.7,
                 verbose: bool = False,
                 greyscale: bool = False,
                 clahe: bool = False,
                 padding: float = 0.00,
                 threshold: float = 0.00,
                 flip_idx: Optional[Tuple[int, ...]] = None,
                 names: Tuple[str, ...] = ('animal_1',),
                 sample_size: Optional[int] = None,
                 bp_id_idx: Optional[Dict[int, Union[Tuple[int], List[int]]]] = None,
                 single_id: Optional[str] = None) -> None:

        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose')
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale')
        check_valid_boolean(value=clahe, source=f'{self.__class__.__name__} clahe')
        check_file_exist_and_readable(file_path=config_path)
        check_float(name=f'{self.__class__.__name__} padding', value=padding, max_value=1.0, min_value=0.0, raise_error=True)
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, max_value=1.0, min_value=0.0)
        check_valid_tuple(x=names, source=f'{self.__class__.__name__} names', valid_dtypes=(str,), minimum_length=1)
        check_if_dir_exists(in_dir=save_dir)
        if flip_idx is not None: check_valid_tuple(x=flip_idx, source=self.__class__.__name__, valid_dtypes=(int,), minimum_length=1)
        self.img_dir, self.lbl_dir = os.path.join(save_dir, 'images'), os.path.join(save_dir, 'labels')
        self.img_train_dir, self.img_val_dir = os.path.join(self.img_dir, 'train'), os.path.join(self.img_dir, 'val')
        self.lbl_train_dir, self.lb_val_dir = os.path.join(self.lbl_dir, 'train'), os.path.join(self.lbl_dir, 'val')
        create_directory(paths=[self.img_train_dir, self.img_val_dir, self.lbl_train_dir, self.lb_val_dir], overwrite=False)
        self.map_path = os.path.join(save_dir, 'map.yaml')
        if single_id is not None: check_str(name=f'{self.__class__.__name__} single_id', value=single_id, raise_error=True)
        if sample_size is not None: check_int(name=f'{self.__class__.__name__} sample', value=sample_size, min_value=1)
        self.config = ConfigReader(config_path=config_path)
        if data_dir is not None:
            check_if_dir_exists(in_dir=data_dir, source=f'{self.__class__.__name__} data_dir')
        else:
            data_dir = self.config.outlier_corrected_dir
        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=[f'.{self.config.file_type}'], raise_error=True, as_dict=True)
        self.video_paths = find_files_of_filetypes_in_directory(directory=self.config.video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, raise_error=True, as_dict=True)
        missing_videos = [x for x in self.data_paths.keys() if x not in self.video_paths.keys()]
        if len(missing_videos) > 0:
            NoDataFoundWarning(msg=f'Data files {missing_videos} do not have corresponding videos in the {self.config.video_dir} directory', source=self.__class__.__name__)
        self.data_w_video = [x for x in self.data_paths.keys() if x in self.video_paths.keys()]
        if len(self.data_w_video) == 0:
            raise NoFilesFoundError(msg=f'None of the data files in {data_dir} have matching videos in the {self.config.video_dir} directory', source=self.__class__.__name__)
        self.sample_size, self.train_size, self.verbose, self.save_dir = sample_size, train_size, verbose, save_dir
        self.greyscale, self.bp_id_idx, self.padding, self.flip_idx = greyscale, bp_id_idx, padding, flip_idx
        self.clahe, self.threshold, self.single_id = clahe, threshold, single_id
        self.names = {0: self.single_id} if self.single_id is not None else {k:v for k, v in enumerate(names)}


    def run(self):
        annotations, timer, body_part_headers = [], SimbaTimer(start=True), []
        for file_cnt, video_name in enumerate(self.data_w_video):
            data = read_df(file_path=self.data_paths[video_name], file_type=self.config.file_type)
            check_valid_dataframe(df=data, source=f'{self.__class__.__name__} {self.data_paths[video_name]}', valid_dtypes=Formats.NUMERIC_DTYPES.value)
            video_path = self.video_paths[video_name]
            check_video_and_data_frm_count_align(video=video_path, data=data, name=self.data_paths[video_name], raise_error=True)
            p_data = data[data.columns[list(data.columns.str.endswith('_p'))]]
            data = data.loc[:, ~data.columns.str.endswith('_p')].reset_index(drop=True)
            data = data.iloc[(p_data[(p_data > self.threshold).all(axis=1)].index)]
            body_part_headers = data.columns
            data['video'], frm_cnt = video_name, len(data)
            if self.sample_size is None:
                video_sample_idx = list(range(0, frm_cnt))
            else:
                video_sample_idx = list(range(0, frm_cnt)) if self.sample_size > frm_cnt else random.sample(list(range(0, frm_cnt)), self.sample_size)
            annotations.append(data.iloc[video_sample_idx].reset_index(drop=False))

        if self.flip_idx is None:
            self.flip_idx = get_yolo_keypoint_flip_idx(x=list(dict.fromkeys([x[:-2] for x in body_part_headers])))

        annotations = pd.concat(annotations, axis=0).reset_index(drop=True)
        video_names = annotations.pop('video').reset_index(drop=True).values
        train_idx = random.sample(list(annotations['index']), int(len(annotations) * self.train_size))
        bp_id_idx = np.array_split(np.array(range(0, int(len(body_part_headers) / 2))), len(self.names.keys()))
        bp_id_idx = [list(x) for x in bp_id_idx]

        for cnt, (idx, idx_data) in enumerate(annotations.iterrows()):
            vid_path = self.video_paths[video_names[cnt]]
            video_meta = get_video_meta_data(video_path=vid_path)
            frm_idx, keypoints = idx_data[0], idx_data.values[1:].reshape(-1, 2)
            mask = (keypoints[:, 0] == 0.0) & (keypoints[:, 1] == 0.0)
            keypoints[mask] = np.nan
            if np.all(np.isnan(keypoints)) or np.all(keypoints == 0.0) or np.all(np.isnan(keypoints) | (keypoints == 0.0)):
                continue
            img_lbl = ''
            if self.verbose:
                print(f'Processing image {cnt + 1}/{len(annotations)}...')
            file_name = f'{video_meta["video_name"]}.{frm_idx}'
            if frm_idx in train_idx:
                img_save_path, lbl_save_path = os.path.join(self.img_train_dir, f'{file_name}.png'), os.path.join(self.lbl_train_dir, f'{file_name}.txt')
            else:
                img_save_path, lbl_save_path = os.path.join(self.img_train_dir, f'{file_name}.png'), os.path.join(self.lb_val_dir, f'{file_name}.txt')
            img = read_frm_of_video(video_path=vid_path, frame_index=frm_idx, greyscale=self.greyscale, clahe=self.clahe)
            img_h, img_w = img.shape[0], img.shape[1]
            keypoints_with_id = {}
            for k, idx in enumerate(bp_id_idx):
                keypoints_with_id[k] = keypoints[idx, :]
            for id, keypoints in keypoints_with_id.items():
                if np.all(np.isnan(keypoints)) or np.all(keypoints == 0.0) or np.all(np.isnan(keypoints) | (keypoints == 0.0)):
                    continue
                visability_col = np.full((keypoints.shape[0], 1), fill_value=2).flatten()
                keypoints = np.insert(keypoints, 2, visability_col, axis=1)
                both_zero = (keypoints[:, 0] == 0) & (keypoints[:, 1] == 0)
                has_nan_or_inf = ~np.isfinite(keypoints[:, 0]) | ~np.isfinite(keypoints[:, 1])
                mask = both_zero | has_nan_or_inf
                keypoints[mask, 2] = 0
                instance_str = f'{id} ' if self.single_id is None else '0 '
                instance_str += keypoint_array_to_yolo_annotation_str(x=keypoints, img_w=img_w, img_h=img_h, padding=self.padding)
                img_lbl += instance_str.strip() + '\n'
                with open(lbl_save_path, mode='wt', encoding='utf-8') as f:
                    f.write(img_lbl)
                cv2.imwrite(img_save_path, img)

        create_yolo_keypoint_yaml(path=self.save_dir, train_path=self.img_train_dir, val_path=self.img_val_dir, names=self.names, save_path=self.map_path, kpt_shape=(len(self.flip_idx), 3), flip_idx=self.flip_idx)
        timer.stop_timer()
        stdout_success(msg=f'YOLO formated data saved in {self.save_dir} directory', source=self.__class__.__name__, elapsed_time=timer.elapsed_time_str)


# SAVE_DIR = r'D:\troubleshooting\mitra\mitra_yolo'
# CONFIG_PATH = r"C:\troubleshooting\mitra\project_folder\project_config.ini"
# runner = SimBA2Yolo(config_path=CONFIG_PATH, save_dir=SAVE_DIR, sample_size=10, verbose=True, names=('animal_1',))
# runner.run()


import os
import random
from typing import Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import cv2

from simba.third_party_label_appenders.transform.utils import (
    create_yolo_keypoint_yaml, get_yolo_keypoint_flip_idx)
from simba.utils.checks import (check_float, check_if_dir_exists, check_int,
                                check_str, check_valid_boolean,
                                check_valid_dataframe, check_valid_tuple)
from simba.utils.enums import Options
from simba.utils.errors import NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (create_directory,
                                    find_files_of_filetypes_in_directory,
                                    get_video_meta_data, read_frm_of_video)
from simba.utils.yolo import keypoint_array_to_yolo_annotation_str


class Sleap2Yolo:
    """
    Convert SLEAP pose estimation CSV data and corresponding videos into YOLO keypoint dataset format.

    .. note::
       This converts SLEAP **inference** data to YOLO keypoints (not SLEAP annotations).

    :param Union[str, os.PathLike] data_dir: Directory path containing SLEAP-generated CSV files with inferred keypoints.
    :param Union[str, os.PathLike] video_dir: Directory path containing corresponding videos from which frames are to be extracted.
    :param Union[str, os.PathLike] save_dir: Output directory where YOLO-formatted images, labels, and map YAML file will be saved. Subdirectories `images/train`, `images/val`, `labels/train`, `labels/val` will be created.
    :param Optional[int] frms_cnt: Number of frames to randomly sample from each video for conversion. If None, all frames are used.
    :param float instance_threshold: Minimum confidence score threshold to filter out low-confidence pose instances. Only instances with `instance.score` >= this threshold are used.
    :param float train_size: Proportion of frames randomly assigned to the training dataset. Value must be between 0.1 and 0.99. Default: 0.7.
    :param bool verbose: If True, prints progress. Default: True.
    :param Tuple[int, ...] flip_idx: Tuple of keypoint indices used for horizontal flip augmentation during training. The tuple defines the order of keypoints after flipping.
    :param Dict[str, int] map_dict: Dictionary mapping class indices to class names. Used for creating the YAML class names mapping file.
    :param float padding: Fractional padding to add around the bounding boxes (relative to image dimensions). Helps to slightly enlarge bounding boxes by this percentage. Default 0.05. E.g., Useful when all body-parts are along animal length.
    :param Optional[str] single_id: If the data contains pose-estimation for multiple individuals, but you want to treat it as examples of a single individual, pass the name of the single individual. Defaults to None, and the YOLO data will be formatted to the number of objects which the H5 data contains.

    :return: None. Results saved in ``save_dir``.

    :example:
    >>> DATA_DIR = r'D:\ares\data\ant\sleap_csv'
    >>> VIDEO_DIR = r'D:\ares\data\ant\sleap_video'
    >>> SAVE_DIR = r"D:\imgs\sleap_csv"
    >>> runner = Sleap2Yolo(data_dir=DATA_DIR, video_dir=VIDEO_DIR, frms_cnt=50, train_size=0.8, instance_threshold=0.9, save_dir=SAVE_DIR, single_id='ant')
    >>> runner.run()

    """
    def __init__(self,
                 data_dir: Union[str, os.PathLike],
                 video_dir: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 frms_cnt: Optional[int] = None,
                 verbose: bool = True,
                 instance_threshold: float = 0,
                 train_size: float = 0.7,
                 flip_idx: Optional[Tuple[int, ...]] = None,
                 names: Optional[Tuple[str, ...]] = None,
                 greyscale: bool = False,
                 clahe: bool = False,
                 padding: float = 0.00,
                 single_id: Optional[str] = None):

        self.data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'], as_dict=True, raise_error=True)
        self.video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=Options.ALL_VIDEO_FORMAT_OPTIONS.value, as_dict=True, raise_error=True)
        missing_video_paths = [x for x in self.video_paths.keys() if x not in self.data_paths.keys()]
        missing_data_paths = [x for x in self.data_paths.keys() if x not in self.video_paths.keys()]
        check_if_dir_exists(in_dir=save_dir)
        self.img_dir, self.lbl_dir = os.path.join(save_dir, 'images'), os.path.join(save_dir, 'labels')
        self.img_train_dir, self.img_val_dir = os.path.join(save_dir, 'images', 'train'), os.path.join(save_dir, 'images', 'val')
        self.lbl_train_dir, self.lb_val_dir = os.path.join(save_dir, 'labels', 'train'), os.path.join(save_dir, 'labels', 'val')
        if flip_idx is not None: check_valid_tuple(x=flip_idx, source=f'{self.__class__.__name__} flip_idx', valid_dtypes=(int,), minimum_length=1)
        if names is not None: check_valid_tuple(x=names, source=f'{self.__class__.__name__} names', valid_dtypes=(str,), minimum_length=1)
        if single_id is not None: check_str(name=f'{self.__class__.__name__} single_id', value=single_id, raise_error=True)
        create_directory(paths=[self.img_train_dir, self.img_val_dir, self.lbl_train_dir, self.lb_val_dir], overwrite=True)
        check_float(name=f'{self.__class__.__name__} instance_threshold', min_value=0.0, max_value=1.0, raise_error=True, value=instance_threshold)
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_boolean(value=greyscale, source=f'{self.__class__.__name__} greyscale', raise_error=True)
        check_float(name=f'{self.__class__.__name__} train_size', value=train_size, max_value=0.99, min_value=0.1)
        check_float(name=f'{self.__class__.__name__} padding', value=padding, max_value=1.0, min_value=0.0, raise_error=True)

        self.map_path = os.path.join(save_dir, 'map.yaml')
        if frms_cnt is not None:
            check_int(name=f'{self.__class__.__name__} frms_cnt', value=frms_cnt, min_value=1, raise_error=True)
        if len(missing_video_paths) > 0:
            raise NoFilesFoundError(msg=f'Video(s) {missing_video_paths} could not be found in {video_dir} directory', source=self.__class__.__name__)
        if len(missing_data_paths) > 0:
            raise NoFilesFoundError(msg=f'CSV data for {missing_data_paths} could not be found in {data_dir} directory', source=self.__class__.__name__)
        self.verbose, self.instance_threshold, self.frms_cnt = verbose, instance_threshold, frms_cnt
        self.names, self.greyscale, self.train_size, self.clahe = names, greyscale, train_size, clahe
        self.padding, self.flip_idx, self.save_dir, self.single_id = padding, flip_idx, save_dir, single_id



    def run(self):
        dfs, timer, bp_cols = [], SimbaTimer(start=True), []
        for file_cnt, (file_name, file_path) in enumerate(self.data_paths.items()):
            df = pd.read_csv(filepath_or_buffer=file_path)
            check_valid_dataframe(df=df, source=self.__class__.__name__, required_fields=['track', 'frame_idx', 'instance.score'])
            df = df if self.instance_threshold is None else df[df['instance.score'] >= self.instance_threshold]
            cord_cols, frame_idx = df.drop(['track', 'frame_idx', 'instance.score'], axis=1), df['frame_idx']
            bp_cols = [x for x in cord_cols if not '.score' in x]
            selected_frms = random.sample(list(frame_idx.unique()), self.frms_cnt) if self.frms_cnt is not None else list(frame_idx.unique())
            df = df[df['frame_idx'].isin(selected_frms)]
            df['video'] = self.video_paths[file_name]
            dfs.append(df)

        dfs = pd.concat(dfs, axis=0)
        unique_tracks_lk = {v: k for k, v in enumerate(dfs['track'].unique())}
        if self.names is not None:
            check_valid_tuple(x=self.names, source=f'{self.__class__.__name__} names', valid_dtypes=(str,), accepted_lengths=(len(list(unique_tracks_lk.keys())),))
        else:
            self.names = tuple([f'animal_{k + 1}' for k in range(len(list(unique_tracks_lk.keys())), )])
        map_dict = {k: v for k, v in enumerate(self.names)} if self.single_id is None else {0: self.single_id}
        dfs['id'] = dfs['frame_idx'].astype(str) + dfs['video'].astype(str)
        train_idx = random.sample(list(dfs['id'].unique()), int(len(dfs['frame_idx'].unique()) * self.train_size))
        if self.flip_idx is None:
            self.flip_idx = get_yolo_keypoint_flip_idx(x=list(dict.fromkeys([x[:-2] for x in bp_cols])))

        for frm_cnt, frm_id in enumerate(dfs['id'].unique()):
            frm_data = dfs[dfs['id'] == frm_id]
            video_path = list(frm_data['video'])[0]
            frm_idx = list(frm_data['frame_idx'])[0]
            video_meta = get_video_meta_data(video_path=video_path)
            if self.verbose:
                print(f'Processing frame: {frm_cnt + 1}/{len(list(dfs["id"].unique()))} ...')
            img = read_frm_of_video(video_path=video_path, frame_index=frm_idx, greyscale=self.greyscale, clahe=self.clahe)
            img_h, img_w = img.shape[0], img.shape[1]
            if list(frm_data['id'])[0] in train_idx:
                img_save_path = os.path.join(self.img_train_dir, f'{video_meta["video_name"]}_{frm_idx}.png')
                lbl_save_path = os.path.join(self.lbl_train_dir, f'{video_meta["video_name"]}_{frm_idx}.txt')
            else:
                img_save_path = os.path.join(self.img_val_dir, f'{video_meta["video_name"]}_{frm_idx}.png')
                lbl_save_path = os.path.join(self.lb_val_dir, f'{video_meta["video_name"]}_{frm_idx}.txt')
            img_lbl = ''
            for track_cnt, (_, track_data) in enumerate(frm_data.iterrows()):
                track_id, keypoints = unique_tracks_lk[track_data['track']], track_data.drop(['track', 'frame_idx', 'instance.score', 'video', 'id']),
                keypoints = keypoints.values.reshape(-1, 3).astype(np.float32)
                keypoints[keypoints[:, 2] != 0.0, 2] = 2
                instance_str = f'{track_id} ' if self.single_id is None else f'0 '
                instance_str += keypoint_array_to_yolo_annotation_str(x=keypoints, img_h=img_h, img_w=img_w, padding=self.padding)
                img_lbl += instance_str
            with open(lbl_save_path, mode='wt', encoding='utf-8') as f:
                f.write(img_lbl)
            cv2.imwrite(img_save_path, img)

        create_yolo_keypoint_yaml(path=self.save_dir, train_path=self.img_train_dir, val_path=self.img_val_dir, names=map_dict, save_path=self.map_path, kpt_shape=(len(self.flip_idx), 3), flip_idx=tuple(self.flip_idx))
        timer.stop_timer()
        stdout_success(msg=f'YOLO formated data saved in {self.save_dir} directory', source=self.__class__.__name__, elapsed_time=timer.elapsed_time_str)


# DATA_DIR = r'D:\ares\data\ant\sleap_csv'
# VIDEO_DIR = r'D:\ares\data\ant\sleap_video'
# SAVE_DIR = r"D:\imgs\sleap_csv"
#
# runner = Sleap2Yolo(data_dir=DATA_DIR, video_dir=VIDEO_DIR, frms_cnt=50, train_size=0.8, instance_threshold=0.9, save_dir=SAVE_DIR, single_id='ant')
# runner.run()


import functools
import multiprocessing
import os
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int,
                                check_valid_boolean, check_valid_dataframe,
                                check_valid_lst, check_valid_tuple)
from simba.utils.data import create_color_palette
from simba.utils.enums import Defaults, Options
from simba.utils.errors import CountError, DataHeaderError, FrameRangeError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    create_directory, find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data,
                                    read_frm_of_video)

FRAME = 'FRAME'
CLASS_ID = 'CLASS_ID'
CONFIDENCE = 'CONFIDENCE'
CLASS_NAME = 'CLASS_NAME'
BOX_CORD_FIELDS = ['X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']
EXPECTED_COLS = [FRAME, CLASS_ID, CLASS_NAME, CONFIDENCE, 'X1', 'Y1', 'X2', 'Y2', 'X3', 'Y3', 'X4', 'Y4']



def _yolo_keypoint_visualizer(frm_ids: np.ndarray,
                              data: pd.DataFrame,
                              threshold: float,
                              video_path: str,
                              save_dir: str,
                              circle_size: int,
                              thickness: int,
                              palettes: dict,
                              bbox: bool,
                              skeleton: list):

    batch_id, frame_rng = frm_ids[0], frm_ids[1]
    start_frm, end_frm, current_frm = frame_rng[0], frame_rng[-1], frame_rng[0]
    video_meta_data = get_video_meta_data(video_path=video_path)
    cap = cv2.VideoCapture(video_path)
    fourcc, font = cv2.VideoWriter_fourcc(*"mp4v"), cv2.FONT_HERSHEY_DUPLEX
    video_save_path = os.path.join(save_dir, f'{batch_id}.mp4')
    video_writer = cv2.VideoWriter(video_save_path, fourcc, video_meta_data["fps"], (video_meta_data["width"], video_meta_data["height"]))
    while current_frm <= end_frm:
        print(f'Processing frame {current_frm}/{video_meta_data["frame_count"]} (batch: {batch_id}, video: {video_meta_data["video_name"]})...')
        img = read_frm_of_video(video_path=video_path, frame_index=current_frm)
        frm_data = data.loc[data[FRAME] == current_frm]
        frm_data = frm_data[frm_data[CONFIDENCE] > threshold]
        for cnt, (row, row_data) in enumerate(frm_data.iterrows()):
            clrs = np.array(palettes[int(row_data[CLASS_ID])]).astype(np.int32)
            bbox_cords = row_data[BOX_CORD_FIELDS].values.astype(np.int32).reshape(-1, 2)
            kp_coords = row_data.drop(EXPECTED_COLS).values.astype(np.int32).reshape(-1, 3)[:, :-1]
            p_cords = row_data.drop(EXPECTED_COLS).values.astype(np.float64).reshape(-1, 3)[:, -1:].flatten()
            clr = tuple(int(c) for c in clrs[0])
            if bbox:
                img = cv2.polylines(img, [bbox_cords], True, clr, thickness=thickness, lineType=cv2.LINE_AA)
            for kp_cnt, kp in enumerate(kp_coords):
                if p_cords[kp_cnt] > threshold:
                    clr = tuple(int(c) for c in clrs[kp_cnt+1])
                    img = cv2.circle(img, (tuple(kp)), circle_size, clr, -1)
            if skeleton is not None:
                for (kp1, kp2) in skeleton:
                    if (row_data[f'{kp1}_P'] > threshold) and (row_data[f'{kp2}_P'] > threshold):
                        pos_1 = np.array([row_data[[f'{kp1}_X', f'{kp1}_Y']].values.astype(np.int32)])
                        pos_2 = np.array([row_data[[f'{kp2}_X', f'{kp2}_Y']].values.astype(np.int32)])
                        img = PlottingMixin().draw_lines_on_img(img=img, start_positions=pos_1, end_positions=pos_2, color=(105,105,105), highlight_endpoint=False, thickness=int(max(1, int(circle_size/2))))
        video_writer.write(img)
        current_frm += 1
    cap.release()
    video_writer.release()
    return batch_id


class YOLOPoseVisualizer():
    """
    Visualizes YOLO-based keypoint pose estimation data on video frames and creates an annotated output video.

    This class takes keypoint data (CSV) and overlays it onto the corresponding video using color-coded keypoints
    and optional filtering. The result is saved as a new annotated video, and supports multicore parallel rendering
    for efficient processing of long videos.

    .. seealso::
       To create YOLO pose data, see :func:`~simba.bounding_box_tools.yolo.yolo_pose_inference.YOLOPoseInference`
       To fit YOLO model, see :func:`~simba.bounding_box_tools.yolo.yolo_fit.FitYolo`

    .. video:: _static/img/YOLOPoseVisualizer.webm
       :width: 900
       :loop:
       :autoplay:

    :param Union[str, os.PathLike] data_path: Path to the CSV file containing keypoint data (output from YOLO pose inference).
    :param Union[str, os.PathLike] video_path: Path to the original input video to overlay keypoints on.
    :param Union[str, os.PathLike] save_dir: Directory to save the resulting annotated video.
    :param Optional[Union[str, Tuple[str, ...]]] palettes: Name of the color palette(s) to use for drawing keypoints. Can be a string or a tuple of strings (e.g., 'Set1', ('Set1', 'Dark2')). Defaults to 'Set1'.
    :param Optional[int] core_cnt: Number of CPU cores to use for parallel rendering. Defaults to -1 (use all available cores).
    :param float threshold: Confidence threshold for visualizing keypoints. Only keypoints with confidence >= threshold are drawn. Defaults to 0.0.
    :param Optional[int] thickness: Thickness of lines connecting keypoints. If None, determined automatically. Defaults to None.
    :param Optional[int] circle_size: Radius of the circles drawn for keypoints. If None, determined automatically based on frame size. Defaults to None.
    :param Optional[bool] verbose: If True, enables logging and progress messages. Defaults to False.


    :example:
    >>> video_path = r"/mnt/c/troubleshooting/mitra/project_folder/videos/501_MA142_Gi_CNO_0521.mp4"
    >>> data_path = "/mnt/c/troubleshooting/mitra/yolo_pose/501_MA142_Gi_CNO_0521.csv"
    >>> kp_vis = YOLOPoseVisualizer(data_path=data_path,
    >>>                            video_path=video_path,
    >>>                            save_dir='/mnt/c/troubleshooting/mitra/yolo_pose/',
    >>>                            core_cnt=18)
    >>> kp_vis.run()
    """



    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 video_path: Union[str, os.PathLike],
                 save_dir: Union[str, os.PathLike],
                 palettes: Optional[Union[str, Tuple[str, ...]]] = 'Set1',
                 core_cnt: Optional[int] = -1,
                 threshold: float = 0.0,
                 thickness: Optional[int] = None,
                 circle_size: Optional[int] = None,
                 verbose: Optional[bool] = False,
                 bbox: Optional[bool] = True,
                 skeleton: List[Tuple[str, str]] = None):

        check_file_exist_and_readable(file_path=data_path)
        self.video_meta_data = get_video_meta_data(video_path=video_path)
        self.data_path, self.video_path = data_path, video_path
        self.video_name = get_fn_ext(filepath=data_path)[1]
        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        if circle_size is None:
            circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(self.video_meta_data['width'], self.video_meta_data['height']), circle_frame_ratio=80)
        else:
            check_int(name=f'{self.__class__.__name__} circle_size', value=circle_size, min_value=1)
        if thickness is None:
            thickness = circle_size
        else:
            check_int(name=f'{self.__class__.__name__} thickness', value=thickness, min_value=1)
        check_float(name=f'{self.__class__.__name__} threshold', value=threshold, min_value=0.0, max_value=1.0)
        self.core_cnt = core_cnt
        if core_cnt == -1 or core_cnt > find_core_cnt()[0]: self.core_cnt = find_core_cnt()[0]
        check_if_dir_exists(in_dir=save_dir)
        check_valid_boolean(value=[verbose], source=f'{self.__class__.__name__} verbose', raise_error=True)
        check_valid_boolean(value=[bbox], source=f'{self.__class__.__name__} bbox', raise_error=True)
        self.data_df = pd.read_csv(self.data_path, index_col=0)
        check_valid_dataframe(df=self.data_df, source=self.__class__.__name__, required_fields=EXPECTED_COLS)
        self.df_frm_cnt = np.unique(self.data_df[FRAME].values).shape[0]
        self.classes = np.unique(self.data_df[CLASS_NAME].values)
        if isinstance(palettes, str):
            palettes = (palettes,)
        else:
            check_valid_tuple(x=palettes, source=f'{self.__class__.__name__} palettes', minimum_length=1, valid_dtypes=(str,), accepted_values=Options.PALETTE_OPTIONS_CATEGORICAL.value)
        if len(self.classes) != len(palettes):
            raise CountError(msg=f'{len(self.classes)} classes detected in {self.data_path}, but {len(palettes)} color palette names passed: {palettes}', source=self.__class__.__name__)
        self.palettes = {}
        for cnt, palette in enumerate(palettes):
            self.palettes[cnt] = create_color_palette(pallete_name=palette, increments=len(self.data_df.columns)-len(EXPECTED_COLS))
        if skeleton is not None:
            check_valid_lst(data=skeleton, source=f'{self.__class__.__name__} skeleton', valid_dtypes=(list, tuple,), min_len=1, raise_error=True)
            for i in skeleton:
                check_valid_tuple(x=i, source=f'{self.__class__.__name__} {i}', accepted_lengths=(2,), valid_dtypes=(str,))
            required_s_cols = [f'{x}_{suffix}' for xs in skeleton for x in xs for suffix in ('X', 'Y', 'P')]
            missing = [x for x in required_s_cols if x not in self.data_df.columns]
            if len(missing) > 0:
                raise DataHeaderError(msg=f'Columns {missing} missing in file {data_path} as passe by skeleton.', source=self.__class__.__name__)

        self.save_dir, self.verbose, self.palette, self.thickness = save_dir, verbose, palettes, thickness
        self.threshold, self.circle_size, self.thickness, self.bbox = threshold, circle_size, thickness, bbox
        self.video_temp_dir = os.path.join(self.save_dir, self.video_name, "temp")
        self.save_path, self.skeleton = os.path.join(self.save_dir, f'{self.video_name}.mp4'), skeleton
        create_directory(paths=self.video_temp_dir)


    def run(self):
        video_timer = SimbaTimer(start=True)
        if self.video_meta_data['frame_count'] != self.df_frm_cnt:
            raise FrameRangeError(msg=f'The bounding boxes contain data for {self.df_frm_cnt} frames, while the video is {self.video_meta_data["frame_count"]} frames ({self.video_meta_data["video_name"]})', source=self.__class__.__name__)
        frm_batches = np.array_split(np.array(list(range(0, self.df_frm_cnt))), self.core_cnt)
        frm_batches = [(i, j) for i, j in enumerate(frm_batches)]
        if self.verbose: print(f'Visualizing video {self.video_meta_data["video_name"]} (frame count: {self.video_meta_data["frame_count"]})...')
        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=Defaults.MAXIMUM_MAX_TASK_PER_CHILD.value) as pool:
            constants = functools.partial(_yolo_keypoint_visualizer,
                                          data=self.data_df,
                                          threshold=self.threshold,
                                          video_path=self.video_path,
                                          save_dir=self.video_temp_dir,
                                          circle_size=self.circle_size,
                                          thickness=self.thickness,
                                          palettes=self.palettes,
                                          bbox=self.bbox,
                                          skeleton=self.skeleton)
            for cnt, result in enumerate(pool.imap(constants, frm_batches, chunksize=1)):
                print(f'Video batch {result+1}/{self.core_cnt} complete...')
        pool.terminate()
        pool.join()
        video_timer.stop_timer()
        concatenate_videos_in_folder(in_folder=self.video_temp_dir, save_path=self.save_path, gpu=True)

        stdout_success(msg=f'YOLO pose video saved at {self.save_path}', source=self.__class__.__name__, elapsed_time=video_timer.elapsed_time_str)


if __name__ == "__main__":
    #video_path = r"D:\cvat_annotations\videos\mp4_20250624155703\s34-drinking.mp4"
    #data_path = r"D:\cvat_annotations\yolo_07032025\out_data\s34-drinking.csv"
    #save_dir = r"D:\cvat_annotations\yolo_07032025\out_video"

    video_path = r"D:\cvat_annotations\videos\mp4_20250624155703\s34-drinking.mp4"

    data_path = r"D:\cvat_annotations\yolo_mdl_07122025\out_data\s34-drinking.csv"
    save_dir = r'D:\cvat_annotations\yolo_mdl_07122025\out_video'
    video_dir = r"D:\cvat_annotations\videos\mp4_20250624155703"
    data_dir = r"D:\cvat_annotations\yolo_mdl_07122025\out_data"
    data_dir = r"D:\platea\platea_videos\videos\yolo_results"
    video_dir = r"D:\platea\platea_videos\videos\videos"
    save_dir = r'D:\platea\platea_videos\videos\yolo_kp_video_out'

    video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=['.mp4'], as_dict=True)
    data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'], as_dict=True)


    skeleton = [('NOSE', 'LEFT_EAR'),
                ('NOSE', 'RIGHT_EAR'),
                ('RIGHT_EAR', 'LEFT_EAR'),
                ('LEFT_EAR', 'LEFT_SIDE'),
                ('RIGHT_EAR', 'RIGHT_SIDE'),
                ('LEFT_SIDE', 'CENTER'),
                ('LEFT_EAR', 'CENTER'),
                ('RIGHT_EAR', 'CENTER'),
                ('CENTER', 'RIGHT_SIDE'),
                ('CENTER', 'TAIL_BASE'),
                ('LEFT_SIDE', 'TAIL_BASE'),
                ('RIGHT_SIDE', 'TAIL_BASE'),
                ('TAIL_BASE', 'TAIL_CENTER'),
                ('TAIL_CENTER', 'TAIL_TIP')]

    skeleton = [('NOSE', 'LEFT_EAR'),
                ('NOSE', 'RIGHT_EAR'),
                ('RIGHT_EAR', 'LEFT_EAR'),
                ('LEFT_EAR', 'LEFT_SIDE'),
                ('RIGHT_EAR', 'RIGHT_SIDE'),
                ('LEFT_SIDE', 'CENTER'),
                ('LEFT_EAR', 'CENTER'),
                ('RIGHT_EAR', 'CENTER'),
                ('CENTER', 'RIGHT_SIDE'),
                ('CENTER', 'TAIL_BASE'),
                ('LEFT_SIDE', 'TAIL_BASE'),
                ('RIGHT_SIDE', 'TAIL_BASE')]

    for video_name, video_path in video_paths.items():
        data_path = data_paths[video_name]
        #kp_vis = YOLOPoseVisualizer(data_path=data_path, video_path=video_path)

        kp_vis = YOLOPoseVisualizer(data_path= data_path,
                                    video_path=video_path,
                                    save_dir=save_dir,
                                    core_cnt=31,
                                    bbox=True,
                                    verbose=True,
                                    skeleton=skeleton,
                                    threshold=0.5,
                                    thickness=3,
                                    circle_size=10)

        kp_vis.run()

    # video_dir = r'D:\cvat_annotations\videos\mp4_20250624155703'
    # data_dir = r'D:\cvat_annotations\yolo_mdl_07102025\out_csv'
    # save_dir = r'D:\cvat_annotations\yolo_mdl_07102025\out_video'
    #
    # video_paths = find_files_of_filetypes_in_directory(directory=video_dir, extensions=['.mp4'], as_dict=True)
    # data_paths = find_files_of_filetypes_in_directory(directory=data_dir, extensions=['.csv'], as_dict=True)
    #
    # for video_name, data_path in data_paths.items():
    #     kp_vis = YOLOPoseVisualizer(data_path= data_path,
    #                                 video_path=video_paths[video_name],
    #                                 save_dir=save_dir,
    #                                 core_cnt=28,
    #                                 bbox=True,
    #                                 verbose=True)
    #
    #     kp_vis.run()


# if __name__ == "__main__":
#     from simba.utils.read_write import find_files_of_filetypes_in_directory
#
#     video_paths = find_files_of_filetypes_in_directory(directory=r'D:\cvat_annotations\videos', extensions=['.mp4'], as_dict=True)
#     data_paths = find_files_of_filetypes_in_directory(directory=r'D:\cvat_annotations\frames\data_results_yolo11m', extensions=['.csv'], as_dict=True)
#     save_dir = r'D:\cvat_annotations\frames\video_results_yolo11m'
#
#     for video_name, video_path in video_paths.items():
#         data_path = data_paths[video_name]
#         kp_vis = YOLOPoseVisualizer(data_path=data_path,
#                                     video_path=video_path,
#                                     save_dir=save_dir,
#                                     core_cnt=18)
#         kp_vis.run()


# video_path = r"/mnt/c/troubleshooting/mitra/project_folder/videos/501_MA142_Gi_CNO_0521.mp4"
# video_path = "/mnt/c/troubleshooting/mitra/project_folder/videos/502_MA141_Gi_CNO_0514.mp4"
# data_path = r"/mnt/c/troubleshooting/mitra/yolo/results/502_MA141_Gi_CNO_0514.csv"
# save_dir = r'/mnt/c/troubleshooting/mitra/yolo/results/'
# kp_vis = YOLOPoseVisualizer(data_path=data_path,
#                             video_path=video_path,
#                             save_dir=save_dir,
#                             core_cnt=18)
#
#
# kp_vis.run()



